"""Class for stream 02 function 23."""

from secsgem.secs.data_items import DSPER, REPGSZ, SVID, TOTSMP, TRID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F23(SecsStreamFunction):
    """Trace initialize.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F23
        {
            TRID: I1/I2/I4/I8/U1/U2/U4/U8/A
            DSPER: A
            TOTSMP: I1/I2/I4/I8/U1/U2/U4/U8/A
            REPGSZ: I1/I2/I4/I8/U1/U2/U4/U8/A
            SVID: [
                DATA: U1/U2/U4/U8/I1/I2/I4/I8/A
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F23({
        ...     "TRID":1,
        ...     "DSPER":'000010',
        ...     "TOTSMP":secsgem.secs.variables.U4(10),
        ...     "REPGSZ":secsgem.secs.variables.U4(1),
        ...     "SVID":[1002004,400210]})
        S2F23 W
          <L [5]
            <I1 1 >
            <A "000010">
            <U4 10 >
            <U4 1 >
            <L [2]
              <U4 1002004 >
              <U4 400210 >
            >
          > .

    Data Items:
        - :class:`TRID <secsgem.secs.data_items.TRID>`
        - :class:`DSPER <secsgem.secs.data_items.DSPER>`
        - :class:`TOTSMP <secsgem.secs.data_items.TOTSMP>`
        - :class:`REPGSZ <secsgem.secs.data_items.REPGSZ>`
        - :class:`SVID <secsgem.secs.data_items.SVID>`

    """

    _stream = 2
    _function = 23

    _data_format = [
        TRID,
        DSPER,
        TOTSMP,
        REPGSZ,
        [SVID]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = True



"""Header for the hsms separate request."""

from .header import HsmsHeader, HsmsSType


class HsmsSeparateReqHeader(HsmsHeader):
    """Header for Separate Request.

    Header for message with SType 9.
    """

    def __init__(self, system: int):
        """Initialize a hsms separate request header.

        :param system: message ID
        :type system: integer

        Example:
            >>> import secsgem.hsms
            >>>
            >>> secsgem.hsms.HsmsSeparateReqHeader(17)
            HsmsSeparateReqHeader({session_id:0xffff, stream:00, function:00, p_type:0x00, s_type:0x09, \
system:0x00000011, require_response:False})
        """
        super().__init__(system, 0xFFFF, 0, 0, False, 0x00, HsmsSType.SEPARATE_REQ)



"""Class for stream 10 function 03."""

from secsgem.secs.data_items import TEXT, TID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS10F03(SecsStreamFunction):
    """terminal single - display.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS10F03
        {
            TID: B[1]
            TEXT: U1/U2/U4/U8/I1/I2/I4/I8/A/B
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS10F03({"TID": 0, "TEXT": "hello!"})
        S10F3
          <L [2]
            <B 0x0>
            <A "hello!">
          > .

    Data Items:
        - :class:`TID <secsgem.secs.data_items.TID>`
        - :class:`TEXT <secsgem.secs.data_items.TEXT>`

    """

    _stream = 10
    _function = 3

    _data_format = [
        TID,
        TEXT
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 01 function 01."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F01(SecsStreamFunction):
    """are you online - request.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F01
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F01()
        S1F1 W .

    """

    _stream = 1
    _function = 1

    _data_format = None

    _to_host = True
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""State machine for connection state."""
from __future__ import annotations

import enum

import secsgem.common


class ConnectionState(enum.Enum):
    """States for connection state machine."""

    NOT_CONNECTED = 0
    CONNECTED = 1
    CONNECTED_NOT_SELECTED = 2
    CONNECTED_SELECTED = 3


class ConnectionStateMachine(secsgem.common.StateMachine):
    """Connection state machine."""

    def __init__(self) -> None:
        """Initialize state machine."""
        super().__init__()

        self.not_connected = secsgem.common.State(
            ConnectionState.NOT_CONNECTED,
            "NOT_CONNECTED",
            initial=True)
        self.connected = secsgem.common.State(
            ConnectionState.CONNECTED,
            "CONNECTED")
        self.connected_not_selected = secsgem.common.State(
            ConnectionState.CONNECTED_NOT_SELECTED,
            "CONNECTED_NOT_SELECTED",
            parent=self.connected)
        self.connected_selected = secsgem.common.State(
            ConnectionState.CONNECTED_SELECTED,
            "CONNECTED_SELECTED",
            parent=self.connected)

        # transition 1
        self._current_state: secsgem.common.State = self.not_connected

        self._transitions: list[secsgem.common.Transition] = [
            secsgem.common.Transition(
                "connect",
                self.not_connected,
                self.connected_not_selected),  # 2
            secsgem.common.Transition(
                "disconnect",
                [self.connected_not_selected, self.connected_selected],
                self.not_connected),  # 3
            secsgem.common.Transition(
                "select",
                self.connected_not_selected,
                self.connected_selected),  # 4
            secsgem.common.Transition(
                "deselect",
                self.connected_selected,
                self.connected_not_selected),  # 5
            secsgem.common.Transition(
                "timeoutT7",
                self.connected_not_selected,
                self.not_connected),  # 6
        ]

    def connect(self) -> None:
        """Perform connect transition."""
        self._perform_transition("connect")

    def disconnect(self) -> None:
        """Perform disconnect transition."""
        self._perform_transition("disconnect")

    def select(self) -> None:
        """Perform select transition."""
        self._perform_transition("select")

    def deselect(self) -> None:
        """Perform deselect transition."""
        self._perform_transition("deselect")



"""Contains objects and functions to create and handle tcp connection."""
from __future__ import annotations

import logging
import select
import threading
import time
import typing

from .connection import Connection
from .helpers import format_hex, is_errorcode_ewouldblock

if typing.TYPE_CHECKING:
    import socket

    from .settings import Settings


class TcpConnection(Connection):
    """Base connection class used for TCP connection types."""

    select_timeout = 0.5
    """Timeout for select calls ."""

    def __init__(self, settings: Settings):
        """Initialize a TCP connection.

        Args:
            settings: protocol and communication settings

        """
        super().__init__(settings)

        self._logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)
        self._bytestream_logger = logging.getLogger("bytestream")

        # connection socket
        self._sock: socket.socket | None = None

        # receiving thread flags
        self._thread_running = False
        self._stop_thread = False

    @property
    def _socket(self) -> socket.socket:
        if self._sock is None:
            raise ConnectionError(f"TCP socket is not connected: {self}")

        return self._sock

    def _serialize_data(self):
        """Return data for serialization.

        Returns:
            data to serialize for this object

        """
        return {
            "connect_mode": self._settings.connect_mode,
            "remoteAddress": self._settings.address,
            "remotePort": self._settings.port,
            "session_id": self._settings.session_id,
            "connected": self._connected}

    def __str__(self):
        """Get the contents of this object as a string."""
        return (
            f"{self._settings.connect_mode} connection to "
            f"{self._settings.address}:{self._settings.port}"
            f" session_id={self._settings.session_id}"
        )

    def _start_receiver(self):
        """Start the thread for receiving and handling incoming messages.

        Will also do the initial Select and Linktest requests.
        """
        # start data receiving thread
        threading.Thread(target=self.__receiver_thread, args=(),
                         name=f"secsgem_tcpConnection_receiver_{self._settings.address}:{self._settings.port}").start()

        # wait until thread is running
        while not self._thread_running:
            pass

    def disconnect(self):
        """Close connection."""
        # return if thread isn't running
        if not self._thread_running:
            return

        # set disconnecting flag to avoid another select
        self._disconnecting = True

        # set flag to stop the thread
        self._stop_thread = True

        # wait until thread stopped
        while self._thread_running:
            pass

        # clear disconnecting flag, no selects coming any more
        self._disconnecting = False

    def send_data(self, data: bytes) -> bool:
        """Send data to the remote host.

        Args:
            data: encoded data.

        Returns:
            True if succeeded, False if failed

        """
        retry = True

        # not sent yet, retry
        while retry:
            # wait until socket is writable
            while not select.select([], [self._socket], [], self.select_timeout)[1]:
                pass

            try:
                # send message
                self._socket.send(data)

                # retry will be cleared if send succeeded
                retry = False
            except OSError as exc:
                if not is_errorcode_ewouldblock(exc.errno):
                    # raise if not EWOULDBLOCK
                    return False
                # it is EWOULDBLOCK, so retry sending

            self._bytestream_logger.debug("> %s", format_hex(data))

        return True

    def __receiver_thread_read_data(self):
        # check if shutdown requested
        while not self._stop_thread:
            # check if data available
            select_result = select.select([self._socket], [], [self._socket], self.select_timeout)

            # check if disconnection was started
            if self._disconnecting:
                time.sleep(0.2)
                continue

            if select_result[0]:
                try:
                    # get data from socket
                    recv_data = self._socket.recv(1024)

                    # check if socket was closed
                    if len(recv_data) == 0:
                        self._connected = False
                        self._stop_thread = True
                        continue

                    self._bytestream_logger.debug("< %s", format_hex(recv_data))

                    # add received data to input buffer
                    self.on_data({"source": self, "data": recv_data})
                except OSError as exc:
                    if not is_errorcode_ewouldblock(exc.errno):
                        raise exc

    def __receiver_thread(self):
        """Thread for receiving incoming data and adding it to the receive buffer."""
        self._thread_running = True

        try:
            self.__receiver_thread_read_data()
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("exception")

        # notify listeners of disconnection
        try:
            self.on_disconnecting({"source": self})
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("ignoring exception for on_connection_before_closed handler")

        # close the socket
        self._socket.close()

        # notify listeners of disconnection
        try:
            self.on_disconnected({"source": self})
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("ignoring exception for on_connection_closed handler")

        # reset all flags
        self._connected = False
        self._thread_running = False
        self._stop_thread = False



"""Connection base function."""
import abc

from .events import Event
from .settings import Settings


class Connection(abc.ABC):
    """Abstract base class for a connection."""

    def __init__(self, settings: Settings) -> None:
        """Initialize connection object."""
        super().__init__()

        self._settings = settings

        self._on_connected = Event()
        self._on_data = Event()
        self._on_disconnecting = Event()
        self._on_disconnected = Event()

        self._connected = False
        self._disconnecting = False

    @property
    def on_connected(self) -> Event:
        """Get the connected event.

        Callbacks to this event are called, when a connection was established.

        """
        return self._on_connected

    @property
    def on_data(self) -> Event:
        """Get the data received event.

        Callbacks to this event are called, when data was received.

        """
        return self._on_data

    @property
    def on_disconnecting(self) -> Event:
        """Get the disconnecting event.

        Callbacks to this event are called, before the connection is separated.

        """
        return self._on_disconnecting

    @property
    def on_disconnected(self) -> Event:
        """Get the disconnect event.

        Callbacks to this event are called, after the connection was separated.

        """
        return self._on_disconnected

    @property
    def connected(self) -> bool:
        """Get the connected flag.

        This flag is True, when a connection is established.

        """
        return self._connected

    @property
    def disconnecting(self) -> bool:
        """Get the disconnecting flag.

        This flag is True, when the connection is about to be separated.

        """
        return self._disconnecting

    @abc.abstractmethod
    def enable(self):
        """Enable the connection.

        Open port and start receiver thread.
        """
        raise NotImplementedError("Connection.enable missing implementation")

    @abc.abstractmethod
    def disable(self):
        """Disable the connection.

        Close port and stop receiver thread.
        """
        raise NotImplementedError("Connection.disable missing implementation")

    @abc.abstractmethod
    def send_data(self, data: bytes) -> bool:
        """Send data to the remote host.

        Args:
            data: encoded data.

        Returns:
            True if succeeded, False if failed

        """
        raise NotImplementedError("Connection.send_data missing implementation")



"""Class for stream 14 function 03."""

from secsgem.secs.data_items import ATTRDATA, ATTRID, OBJID, OBJSPEC, OBJTYPE
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS14F03(SecsStreamFunction):
    """SetAttr request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS14F03
        {
            OBJSPEC: A
            OBJTYPE: U1/U2/U4/U8/A
            OBJID: [
                DATA: U1/U2/U4/U8/A
                ...
            ]
            ATTRIBS: [
                {
                    ATTRID: U1/U2/U4/U8/A
                    ATTRDATA: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS14F03({
        ...       "OBJSPEC": '',
        ...       "OBJTYPE": 'StripMap',
        ...       "OBJID": ['MAP001'],
        ...       "ATTRIBS": [{
        ...           "ATTRID": "CellStatus",
        ...           "ATTRDATA": "3"}]})
        S14F3 W
          <L [4]
            <A>
            <A "StripMap">
            <L [1]
              <A "MAP001">
            >
            <L [1]
              <L [2]
                <A "CellStatus">
                <A "3">
              >
            >
          > .

    Data Items:
        - :class:`OBJSPEC <secsgem.secs.data_items.OBJSPEC>`
        - :class:`OBJTYPE <secsgem.secs.data_items.OBJTYPE>`
        - :class:`OBJID <secsgem.secs.data_items.OBJID>`
        - :class:`ATTRID <secsgem.secs.data_items.ATTRID>`
        - :class:`ATTRDATA <secsgem.secs.data_items.ATTRDATA>`

    """

    _stream = 14
    _function = 3

    _data_format = [
        OBJSPEC,
        OBJTYPE,
        [OBJID],
        [
            [
                "ATTRIBS",
                ATTRID,
                ATTRDATA
            ]
        ]
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



from secsgem.secs.functions.base import SecsStreamFunction
from secsgem.secs.data_items.tiack import TIACK


class SecsS02F32(SecsStreamFunction):
    _stream = 2
    _function = 32

    _data_format = TIACK

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 01 function 03."""

from secsgem.secs.data_items import SVID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F03(SecsStreamFunction):
    """Selected equipment status - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F03
        [
            SVID: U1/U2/U4/U8/I1/I2/I4/I8/A
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F03([1, "1337", 12])
        S1F3 W
          <L [3]
            <U1 1 >
            <A "1337">
            <U1 12 >
          > .

    Data Items:
        - :class:`SVID <secsgem.secs.data_items.SVID>`

    """

    _stream = 1
    _function = 3

    _data_format = [SVID]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""SECS variable base type."""
from __future__ import annotations


class Base:
    """Base class for SECS variables.

    Due to the python types, wrapper classes for variables are required.
    If constructor is called with Base or subclass only the value is copied.
    """

    format_code = -1
    preferred_types: list[type] | None

    def __init__(self, value=None):
        """Initialize a secs variable."""
        self.value = value

    def __call__(self, *args, **kwargs):
        return self

    def set(self, value):
        """Set the internal value to the provided value.

        :param value: new value
        :type value: various
        """
        raise NotImplementedError("Function set not implemented on " + self.__class__.__name__)

    def encode_item_header(self, length):
        """Encode item header depending on the number of length bytes required.

        :param length: number of bytes in data
        :type length: integer
        :returns: encoded item header bytes
        :rtype: string
        """
        if length < 0:
            raise ValueError(f"Encoding {self.__class__.__name__} not possible, data length too small {length}")
        if length > 0xFFFFFF:
            raise ValueError(f"Encoding {self.__class__.__name__} not possible, data length too big {length}")

        if length > 0xFFFF:
            length_bytes = 3
            format_byte = (self.format_code << 2) | length_bytes
            return bytes(bytearray((format_byte, (length & 0xFF0000) >> 16, (length & 0x00FF00) >> 8,
                                    (length & 0x0000FF))))
        if length > 0xFF:
            length_bytes = 2
            format_byte = (self.format_code << 2) | length_bytes
            return bytes(bytearray((format_byte, (length & 0x00FF00) >> 8, (length & 0x0000FF))))

        length_bytes = 1
        format_byte = (self.format_code << 2) | length_bytes
        return bytes(bytearray((format_byte, (length & 0x0000FF))))

    def decode_item_header(self, data, text_pos=0):
        """Encode item header depending on the number of length bytes required.

        :param data: encoded data
        :type data: string
        :param text_pos: start of item header in data
        :type text_pos: integer
        :returns: start position for next item, format code, length item of data
        :rtype: (integer, integer, integer)
        """
        if len(data) == 0:
            raise ValueError(f"Decoding for {self.__class__.__name__} without any text")

        # parse format byte
        format_byte = bytearray(data)[text_pos]

        format_code = (format_byte & 0b11111100) >> 2
        length_bytes = format_byte & 0b00000011

        text_pos += 1

        # read 1-3 length bytes
        length = 0
        for _ in range(length_bytes):
            length <<= 8
            length += bytearray(data)[text_pos]

            text_pos += 1

        if 0 <= self.format_code != format_code:
            raise ValueError(
                f"Decoding data for {self.__class__.__name__} ({self.format_code}) has invalid format {format_code}")

        return text_pos, format_code, length

    @property
    def is_dynamic(self) -> bool:
        """Check if this instance is Dynamic or derived."""
        return False

    @property
    def preferred_type(self):
        """Get the preferred type for this variable."""
        return self.preferred_types[0]



"""V data item."""
from .. import variables
from .base import DataItemBase


class V(DataItemBase):  # pylint: disable=invalid-name
    """Variable data.

    :Types:
       - :class:`Array <secsgem.secs.variables.Array>`
       - :class:`Boolean <secsgem.secs.variables.Boolean>`
       - :class:`U1 <secsgem.secs.variables.U1>`
       - :class:`U2 <secsgem.secs.variables.U2>`
       - :class:`U4 <secsgem.secs.variables.U4>`
       - :class:`U8 <secsgem.secs.variables.U8>`
       - :class:`I1 <secsgem.secs.variables.I1>`
       - :class:`I2 <secsgem.secs.variables.I2>`
       - :class:`I4 <secsgem.secs.variables.I4>`
       - :class:`I8 <secsgem.secs.variables.I8>`
       - :class:`F4 <secsgem.secs.variables.F4>`
       - :class:`F8 <secsgem.secs.variables.F8>`
       - :class:`String <secsgem.secs.variables.String>`
       - :class:`Binary <secsgem.secs.variables.Binary>`

    **Used In Function**
        - :class:`SecsS06F11 <secsgem.secs.functions.SecsS06F11>`
        - :class:`SecsS06F16 <secsgem.secs.functions.SecsS06F16>`
        - :class:`SecsS06F20 <secsgem.secs.functions.SecsS06F20>`
        - :class:`SecsS06F22 <secsgem.secs.functions.SecsS06F22>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.Array,
        variables.Boolean,
        variables.U1,
        variables.U2,
        variables.U4,
        variables.U8,
        variables.I1,
        variables.I2,
        variables.I4,
        variables.I8,
        variables.F4,
        variables.F8,
        variables.String,
        variables.Binary
    ]



"""Class for stream 10 function 04."""

from secsgem.secs.data_items import ACKC10
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS10F04(SecsStreamFunction):
    """terminal single - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS10F04
        ACKC10: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS10F04(secsgem.secs.data_items.ACKC10.TERMINAL_NOT_AVAILABLE)
        S10F4
          <B 0x2> .

    Data Items:
        - :class:`ACKC10 <secsgem.secs.data_items.ACKC10>`

    """

    _stream = 10
    _function = 4

    _data_format = ACKC10

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 12 function 06."""

from secsgem.secs.data_items import GRNT1
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F06(SecsStreamFunction):
    """map transmit - grant.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F06
        GRNT1: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F06(secsgem.secs.data_items.GRNT1.MATERIALID_UNKNOWN)
        S12F6
          <B 0x5> .

    Data Items:
        - :class:`GRNT1 <secsgem.secs.data_items.GRNT1>`

    """

    _stream = 12
    _function = 6

    _data_format = GRNT1

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 02 function 30."""

from secsgem.secs.data_items import ECDEF, ECID, ECMAX, ECMIN, ECNAME, UNITS
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F30(SecsStreamFunction):
    """equipment constant namelist.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F30
        [
            {
                ECID: U1/U2/U4/U8/I1/I2/I4/I8/A
                ECNAME: A
                ECMIN: BOOLEAN/I8/I1/I2/I4/F8/F4/U8/U1/U2/U4/A/B
                ECMAX: BOOLEAN/I8/I1/I2/I4/F8/F4/U8/U1/U2/U4/A/B
                ECDEF: BOOLEAN/I8/I1/I2/I4/F8/F4/U8/U1/U2/U4/A/B
                UNITS: A
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F30([
        ...     {"ECID": 1,
        ...       "ECNAME": "EC1",
        ...       "ECMIN": secsgem.secs.variables.U1(0),
        ...       "ECMAX": secsgem.secs.variables.U1(100),
        ...       "ECDEF": secsgem.secs.variables.U1(50),
        ...       "UNITS": "mm"},
        ...     {"ECID": 1337,
        ...       "ECNAME": "EC2",
        ...       "ECMIN": "",
        ...       "ECMAX": "",
        ...       "ECDEF": "",
        ...       "UNITS": ""}])
        S2F30
          <L [2]
            <L [6]
              <U1 1 >
              <A "EC1">
              <U1 0 >
              <U1 100 >
              <U1 50 >
              <A "mm">
            >
            <L [6]
              <U2 1337 >
              <A "EC2">
              <A>
              <A>
              <A>
              <A>
            >
          > .

    Data Items:
        - :class:`ECID <secsgem.secs.data_items.ECID>`
        - :class:`ECNAME <secsgem.secs.data_items.ECNAME>`
        - :class:`ECMIN <secsgem.secs.data_items.ECMIN>`
        - :class:`ECMAX <secsgem.secs.data_items.ECMAX>`
        - :class:`ECDEF <secsgem.secs.data_items.ECDEF>`
        - :class:`UNITS <secsgem.secs.data_items.UNITS>`

    """

    _stream = 2
    _function = 30

    _data_format = [
        [
            ECID,
            ECNAME,
            ECMIN,
            ECMAX,
            ECDEF,
            UNITS
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""DVVALNAME data item."""
from .. import variables
from .base import DataItemBase


class DVVALNAME(DataItemBase):
    """Data value name.

    :Type: :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS01F22 <secsgem.secs.functions.SecsS01F22>`

    """

    __type__ = variables.String



"""Class for stream 06 function 15."""

from secsgem.secs.data_items import CEID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F15(SecsStreamFunction):
    """event report request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F15
        CEID: U1/U2/U4/U8/I1/I2/I4/I8/A

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F15(1337)
        S6F15 W
          <U2 1337 > .

    Data Items:
        - :class:`CEID <secsgem.secs.data_items.CEID>`

    """

    _stream = 6
    _function = 15

    _data_format = CEID

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 02 function 26."""

from secsgem.secs.data_items import ABS
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F26(SecsStreamFunction):
    """Loopback diagnostic data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F26
        ABS: B

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F26("Text")
        S2F26
          <B 0x54 0x65 0x78 0x74> .

    Data Items:
        - :class:`ABS <secsgem.secs.data_items.ABS>`

    """

    _stream = 2
    _function = 26

    _data_format = ABS

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 05 function 07."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F07(SecsStreamFunction):
    """list enabled alarms - request.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F07
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F07()
        S5F7 W .

    """

    _stream = 5
    _function = 7

    _data_format = None

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""STRP data item."""
from .. import variables
from .base import DataItemBase


class STRP(DataItemBase):
    """Starting position.

    :Types:
       - :class:`I1 <secsgem.secs.variables.I1>`
       - :class:`I2 <secsgem.secs.variables.I2>`
       - :class:`I4 <secsgem.secs.variables.I4>`
       - :class:`I8 <secsgem.secs.variables.I8>`
    :Length: 2

    **Used In Function**
        - :class:`SecsS12F09 <secsgem.secs.functions.SecsS12F09>`
        - :class:`SecsS12F16 <secsgem.secs.functions.SecsS12F16>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.I1,
        variables.I2,
        variables.I4,
        variables.I8
    ]
    __count__ = 2



"""FNLOC data item."""
from .. import variables
from .base import DataItemBase


class FNLOC(DataItemBase):
    """Flat/notch location.

    In degrees from the bottom CW. (Bottom equals zero degrees.) Zero length indicates not used.

    :Type: :class:`U2 <secsgem.secs.variables.U2>`

    **Used In Function**
        - :class:`SecsS12F01 <secsgem.secs.functions.SecsS12F01>`
        - :class:`SecsS12F03 <secsgem.secs.functions.SecsS12F03>`
        - :class:`SecsS12F04 <secsgem.secs.functions.SecsS12F04>`

    """

    __type__ = variables.U2



"""Class for stream 01 function 14."""

from secsgem.secs.data_items import COMMACK, MDLN
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F14(SecsStreamFunction):
    """establish communication - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F14
        {
            COMMACK: B[1]
            MDLN: [
                DATA: A[20]
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F14({"COMMACK": secsgem.secs.data_items.COMMACK.ACCEPTED,
        ...     "MDLN": ["secsgem", "0.0.6"]})
        S1F14
          <L [2]
            <B 0x0>
            <L [2]
              <A "secsgem">
              <A "0.0.6">
            >
          > .

    Data Items:
        - :class:`COMMACK <secsgem.secs.data_items.COMMACK>`
        - :class:`MDLN <secsgem.secs.data_items.MDLN>`

    .. caution::

        This Stream/function has different structures depending on the source.
        See structure definition below for details.
        Be sure to fill the array accordingly.

    **Structure E->H**::

        {
            COMMACK: B[1]
            DATA: {
                MDLN: A[20]
                SOFTREV: A[20]
            }
        }

    **Structure H->E**::

        {
            COMMACK: B[1]
            DATA: []
        }

    """

    _stream = 1
    _function = 14

    _data_format = [
        COMMACK,
        [MDLN]
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 07 function 00."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F00(SecsStreamFunction):
    """abort transaction stream 7.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F00
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F00()
        S7F0 .

    """

    _stream = 7
    _function = 0

    _data_format = None

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 01 function 02."""

from secsgem.secs.data_items import MDLN
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F02(SecsStreamFunction):
    """on line data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F02
        [
            MDLN: A[20]
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F02(["secsgem", "0.0.6"]) # E->H
        S1F2
          <L [2]
            <A "secsgem">
            <A "0.0.6">
          > .
        >>> secsgem.secs.functions.SecsS01F02() # H->E
        S1F2
          <L> .

    Data Items:
        - :class:`MDLN <secsgem.secs.data_items.MDLN>`

    .. caution::

        This Stream/function has different structures depending on the source.
        If it is sent from the eqipment side it has the structure below, if it
        is sent from the host it is an empty list.
        Be sure to fill the array accordingly.

    **Structure E->H**::

        {
            MDLN: A[20]
            SOFTREV: A[20]
        }

    """

    _stream = 1
    _function = 2

    _data_format = [MDLN]

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""ERRTEXT data item."""
from .. import variables
from .base import DataItemBase


class ERRTEXT(DataItemBase):
    """Error description for error code.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 120

    **Used In Function**
        - :class:`SecsS03F18 <secsgem.secs.functions.SecsS03F18>`
        - :class:`SecsS05F14 <secsgem.secs.functions.SecsS05F14>`
        - :class:`SecsS05F15 <secsgem.secs.functions.SecsS05F15>`
        - :class:`SecsS05F18 <secsgem.secs.functions.SecsS05F18>`
        - :class:`SecsS14F02 <secsgem.secs.functions.SecsS14F02>`
        - :class:`SecsS14F04 <secsgem.secs.functions.SecsS14F04>`

    """

    __type__ = variables.String
    __count__ = 120



"""Class for stream 01 function 12."""

from secsgem.secs.data_items import SVID, SVNAME, UNITS
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F12(SecsStreamFunction):
    """status variable namelist - reply.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F12
        [
            {
                SVID: U1/U2/U4/U8/I1/I2/I4/I8/A
                SVNAME: A
                UNITS: A
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F12([{"SVID": 1, "SVNAME": "SV1", "UNITS": "mm"},
        ...     {"SVID": 1337, "SVNAME": "SV2", "UNITS": ""}])
        S1F12
          <L [2]
            <L [3]
              <U1 1 >
              <A "SV1">
              <A "mm">
            >
            <L [3]
              <U2 1337 >
              <A "SV2">
              <A>
            >
          > .

    Data Items:
        - :class:`SVID <secsgem.secs.data_items.SVID>`
        - :class:`SVNAME <secsgem.secs.data_items.SVNAME>`
        - :class:`UNITS <secsgem.secs.data_items.UNITS>`

    """

    _stream = 1
    _function = 12

    _data_format = [
        [
            SVID,
            SVNAME,
            UNITS
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""ECID data item."""
from .. import variables
from .base import DataItemBase


class ECID(DataItemBase):
    """Equipment constant ID.

    :Types:
       - :class:`U1 <secsgem.secs.variables.U1>`
       - :class:`U2 <secsgem.secs.variables.U2>`
       - :class:`U4 <secsgem.secs.variables.U4>`
       - :class:`U8 <secsgem.secs.variables.U8>`
       - :class:`I1 <secsgem.secs.variables.I1>`
       - :class:`I2 <secsgem.secs.variables.I2>`
       - :class:`I4 <secsgem.secs.variables.I4>`
       - :class:`I8 <secsgem.secs.variables.I8>`
       - :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS02F13 <secsgem.secs.functions.SecsS02F13>`
        - :class:`SecsS02F15 <secsgem.secs.functions.SecsS02F15>`
        - :class:`SecsS02F29 <secsgem.secs.functions.SecsS02F29>`
        - :class:`SecsS02F30 <secsgem.secs.functions.SecsS02F30>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.U1,
        variables.U2,
        variables.U4,
        variables.U8,
        variables.I1,
        variables.I2,
        variables.I4,
        variables.I8,
        variables.String
    ]



"""Remote Control capability."""
from __future__ import annotations

import secsgem.secs

from .capability import Capability
from .collection_event import CollectionEventId
from .handler import GemHandler
from .remote_command import RemoteCommand, RemoteCommandId


class RemoteControlCapability(GemHandler, Capability):
    """Remote Control capability on GEM equipment."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self._remote_commands: dict[int | str | RemoteCommandId, RemoteCommand] = {
            RemoteCommandId.START.value: RemoteCommand(
                RemoteCommandId.START,
                "Start",
                [],
                CollectionEventId.CMD_START_DONE.value),
            RemoteCommandId.STOP.value: RemoteCommand(
                RemoteCommandId.STOP,
                "Stop",
                [],
                CollectionEventId.CMD_STOP_DONE.value),
        }

    @property
    def remote_commands(self) -> dict[int | str | RemoteCommandId, RemoteCommand]:
        """Get list of the remote commands.

        Returns:
            Remote command list

        """
        return self._remote_commands

    def _on_s02f41(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 2, Function 41, host command send.

        The remote command handing differs from usual stream function handling, because we send the ack with later
        completion first.
        Then we run the actual remote command callback and signal success with the matching collection event.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        rcmd_name = function.RCMD.get()
        pre_check_func = getattr(self, f"{rcmd_name}_pre_check", None)
        if pre_check_func:
            if not pre_check_func():
                return self.stream_function(2, 42)({
                    "HCACK": secsgem.secs.data_items.HCACK.CANT_PERFORM_NOW, "PARAMS": []
                })

        rcmd_callback_name = "rcmd_" + rcmd_name

        if rcmd_name not in self._remote_commands:
            self._logger.info("remote command %s not registered", rcmd_name)
            return self.stream_function(2, 42)({"HCACK": secsgem.secs.data_items.HCACK.INVALID_COMMAND, "PARAMS": []})

        if rcmd_callback_name not in self._callback_handler:
            self._logger.warning("callback for remote command %s not available", rcmd_name)
            return self.stream_function(2, 42)({"HCACK": secsgem.secs.data_items.HCACK.INVALID_COMMAND, "PARAMS": []})

        for param in function.PARAMS:
            if param.CPNAME.get() not in self._remote_commands[rcmd_name].params:
                self._logger.warning("parameter %s for remote command %s not available", param.CPNAME.get(), rcmd_name)
                return self.stream_function(2, 42)({
                    "HCACK": secsgem.secs.data_items.HCACK.PARAMETER_INVALID, "PARAMS": []
                })

        self.send_response(self.stream_function(2, 42)({
            "HCACK": secsgem.secs.data_items.HCACK.ACK_FINISH_LATER, "PARAMS": []
        }), message.header.system)

        callback = getattr(self._callback_handler, rcmd_callback_name)

        kwargs = {}
        for param in function.PARAMS.get():
            kwargs[param["CPNAME"]] = param["CPVAL"]

        callback(**kwargs)

        self.trigger_collection_events([self._remote_commands[rcmd_name].ce_finished])

        return None

    def _on_rcmd_START(self):  # noqa: N802 pylint: disable=invalid-name
        self._logger.warning("remote command START not implemented, this is required for GEM compliance")

    def _on_rcmd_STOP(self):  # noqa: N802 pylint: disable=invalid-name
        self._logger.warning("remote command STOP not implemented, this is required for GEM compliance")



"""Class for stream 12 function 03."""

from secsgem.secs.data_items import BCEQU, FFROT, FNLOC, IDTYP, MAPFT, MID, NULBC, ORLOC, PRAXI
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS12F03(SecsStreamFunction):
    """map setup data - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F03
        {
            MID: A/B[80]
            IDTYP: B[1]
            MAPFT: B[1]
            FNLOC: U2
            FFROT: U2
            ORLOC: B[1]
            PRAXI: B[1]
            BCEQU: U1/A
            NULBC: U1/A
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS12F03({
        ...     "MID": "materialID",
        ...     "IDTYP": secsgem.secs.data_items.IDTYP.WAFER_CASSETTE,
        ...     "MAPFT": secsgem.secs.data_items.MAPFT.ROW,
        ...     "FNLOC": 0,
        ...     "FFROT": 0,
        ...     "ORLOC": secsgem.secs.data_items.ORLOC.LOWER_LEFT,
        ...     "PRAXI": secsgem.secs.data_items.PRAXI.COLS_LEFT_INCR,
        ...     "BCEQU": [1, 3, 5, 7],
        ...     "NULBC": "{x}"})
        S12F3 W
          <L [9]
            <A "materialID">
            <B 0x1>
            <B 0x0>
            <U2 0 >
            <U2 0 >
            <B 0x3>
            <B 0x4>
            <U1 1 3 5 7 >
            <A "{x}">
          > .

    Data Items:
        - :class:`MID <secsgem.secs.data_items.MID>`
        - :class:`IDTYP <secsgem.secs.data_items.IDTYP>`
        - :class:`MAPFT <secsgem.secs.data_items.MAPFT>`
        - :class:`FNLOC <secsgem.secs.data_items.FNLOC>`
        - :class:`FFROT <secsgem.secs.data_items.FFROT>`
        - :class:`ORLOC <secsgem.secs.data_items.ORLOC>`
        - :class:`PRAXI <secsgem.secs.data_items.PRAXI>`
        - :class:`BCEQU <secsgem.secs.data_items.BCEQU>`
        - :class:`NULBC <secsgem.secs.data_items.NULBC>`

    """

    _stream = 12
    _function = 3

    _data_format = [
        MID,
        IDTYP,
        MAPFT,
        FNLOC,
        FFROT,
        ORLOC,
        PRAXI,
        BCEQU,
        NULBC
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""SECS numeric variable base type."""
from __future__ import annotations

import struct

from .base import Base


class BaseNumber(Base):
    """Secs base type for numeric data."""

    format_code = 0
    text_code = ""
    _base_type: type[int] | type[float] = int
    _min: int | float = 0
    _max: int | float = 0
    _bytes = 0
    _struct_code = ""

    def __init__(self, value=None, count=-1):
        """Initialize a numeric secs variable.

        :param value: initial value
        :type value: list/integer/float
        :param count: number of items this value
        :type count: integer
        """
        super().__init__()

        self.value = []
        self.count = count
        if value is not None:
            self.set(value)

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        if len(self.value) == 0:
            return f"<{self.text_code}>"

        data = " ".join([str(value) for value in self.value])

        return f"<{self.text_code} {data} >"

    def __len__(self):
        """Get the length."""
        return len(self.value)

    def __getitem__(self, key):
        """Get an item using the indexer operator."""
        return self.value[key]

    def __setitem__(self, key, item):
        """Set an item using the indexer operator."""
        self.value[key] = item

    def __eq__(self, other):
        """Check equality with other object."""
        if isinstance(other, Base):
            if other.is_dynamic:
                return other.value.value == self.value
            return other.value == self.value
        if isinstance(other, list):
            return other == self.value
        return [other] == self.value

    def __hash__(self):
        """Get data item for hashing."""
        return hash(str(self.value))

    def _check_single_item_support(self, value):
        if isinstance(value, float) and self._base_type is int:
            return False

        if isinstance(value, bool):
            return True

        if isinstance(value, (int, float)):
            if value < self._min or value > self._max:
                return False
            return True

        if isinstance(value, (bytes, str)):
            return self._check_single_item_support_str(value)

        return False

    def _check_single_item_support_str(self, value) -> bool:
        try:
            val = self._base_type(value)
        except ValueError:
            return False
        if val < self._min or val > self._max:
            return False
        return True

    def supports_value(self, value) -> bool:
        """Check if the current instance supports the provided value.

        :param value: value to test
        :type value: any
        """
        if isinstance(value, (list, tuple)):
            return self._supports_value_list(value)
        if isinstance(value, bytearray):
            return self._supports_value_bytearray(value)
        return self._check_single_item_support(value)

    def _supports_value_list(self, value) -> bool:
        if 0 <= self.count < len(value):
            return False
        return all(self._check_single_item_support(item) for item in value)

    def _supports_value_bytearray(self, value) -> bool:
        if 0 <= self.count < len(value):
            return False
        return all(not (item < self._min or item > self._max) for item in value)

    def set(self, value):
        """Set the internal value to the provided value.

        :param value: new value
        :type value: list/integer/float
        """
        if isinstance(value, float) and self._base_type is int:
            raise ValueError(f"Invalid value {value}")

        if isinstance(value, (list, tuple)):
            self._set_list(value)
        elif isinstance(value, bytearray):
            self._set_bytearray(value)
        else:
            new_value = self._base_type(value)

            if new_value < self._min or new_value > self._max:
                raise ValueError(f"Invalid value {value}")

            self.value = [new_value]

    def _set_list(self, value):
        if 0 <= self.count < len(value):
            raise ValueError(f"Value longer than {self.count} chars")

        new_list = []
        for item in value:
            item = self._base_type(item)
            if item < self._min or item > self._max:
                raise ValueError(f"Invalid value {item}")

            new_list.append(item)
        self.value = new_list

    def _set_bytearray(self, value):
        if 0 <= self.count < len(value):
            raise ValueError(f"Value longer than {self.count} chars")

        new_list = []
        for item in value:
            if item < self._min or item > self._max:
                raise ValueError(f"Invalid value {item}")
            new_list.append(item)
        self.value = new_list

    def get(self):
        """Return the internal value.

        :returns: internal value
        :rtype: list/integer/float
        """
        if len(self.value) == 1:
            return self.value[0]

        return self.value

    def encode(self):
        """Encode the value to secs data.

        :returns: encoded data bytes
        :rtype: string
        """
        result = self.encode_item_header(len(self.value) * self._bytes)

        for value in self.value:
            result += struct.pack(f">{self._struct_code}", value)

        return result

    def decode(self, data, start=0):
        """Decode the secs byte data to the value.

        :param data: encoded data bytes
        :type data: string
        :param start: start position of value the data
        :type start: integer
        :returns: new start position
        :rtype: integer
        """
        (text_pos, _, length) = self.decode_item_header(data, start)

        result = []

        for _ in range(length // self._bytes):
            result_text = data[text_pos:text_pos + self._bytes]

            if len(result_text) != self._bytes:
                raise ValueError(
                    f"No enough data found for {self.__class__.__name__} with length {length} at position {start} ")

            result.append(struct.unpack(f">{self._struct_code}", result_text)[0])

            text_pos += self._bytes

        self.set(result)

        return text_pos



"""SECS 8 byte signed integer variable type."""

from .base_number import BaseNumber


class I8(BaseNumber):
    """Secs type for 8 byte signed data.

    :param value: initial value
    :type value: list/integer
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o30
    text_code = "I8"
    _base_type = int
    _min = -9223372036854775808
    _max = 9223372036854775807
    _bytes = 8
    _struct_code = "q"
    preferred_types = [int]



"""Class for stream 14 function 04."""

from secsgem.secs.data_items import ATTRDATA, ATTRID, ERRCODE, ERRTEXT, OBJACK, OBJID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS14F04(SecsStreamFunction):
    """SetAttr data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS14F04
        {
            DATA: [
                {
                    OBJID: U1/U2/U4/U8/A
                    ATTRIBS: [
                        {
                            ATTRID: U1/U2/U4/U8/A
                            ATTRDATA: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
                        }
                        ...
                    ]
                }
                ...
            ]
            ERRORS: {
                OBJACK: U1[1]
                ERROR: [
                    {
                        ERRCODE: I1/I2/I4/I8
                        ERRTEXT: A[120]
                    }
                    ...
                ]
            }
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS14F04({
        ...     "DATA": [{
        ...         "OBJID": "MAP001",
        ...         "ATTRIBS": [
        ...             {"ATTRID": "OriginLocation", "ATTRDATA": "0"},
        ...             {"ATTRID": "Rows", "ATTRDATA": 4},
        ...             {"ATTRID": "Columns", "ATTRDATA": 4},
        ...             {"ATTRID": "CellStatus", "ATTRDATA": 6},
        ...             {"ATTRID": "LotID", "ATTRDATA":"LOT001"}]}],
        ...         "ERRORS": {"OBJACK": 0}})
        S14F4
          <L [2]
            <L [1]
              <L [2]
                <A "MAP001">
                <L [5]
                  <L [2]
                    <A "OriginLocation">
                    <A "0">
                  >
                  <L [2]
                    <A "Rows">
                    <U1 4 >
                  >
                  <L [2]
                    <A "Columns">
                    <U1 4 >
                  >
                  <L [2]
                    <A "CellStatus">
                    <U1 6 >
                  >
                  <L [2]
                    <A "LotID">
                    <A "LOT001">
                  >
                >
              >
            >
            <L [2]
              <U1 0 >
              <L>
            >
          > .

    Data Items:
        - :class:`OBJID <secsgem.secs.data_items.OBJID>`
        - :class:`ATTRID <secsgem.secs.data_items.ATTRID>`
        - :class:`ATTRDATA <secsgem.secs.data_items.ATTRDATA>`
        - :class:`OBJACK <secsgem.secs.data_items.OBJACK>`
        - :class:`ERRCODE <secsgem.secs.data_items.ERRCODE>`
        - :class:`ERRTEXT <secsgem.secs.data_items.ERRTEXT>`

    """

    _stream = 14
    _function = 4

    _data_format = [
        [
            [
                OBJID,
                [
                    [
                        "ATTRIBS",
                        ATTRID,
                        ATTRDATA
                    ]
                ]
            ]
        ],
        [
            "ERRORS",
            OBJACK,
            [
                [
                    "ERROR",
                    ERRCODE,
                    ERRTEXT
                ]
            ]
        ]
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Secs I over TCP settings class."""
from __future__ import annotations

import enum

import secsgem.common
import secsgem.secsi


class SecsITcpConnectMode(enum.Enum):
    """Secs I over TCP connect mode (client or server)."""

    CLIENT = 1
    SERVER = 2

    def __repr__(self) -> str:
        """String representation of object."""
        return "Client" if self == self.CLIENT else "Server"


class SecsITcpSettings(secsgem.secsi.SecsISettings):
    """Settings for Secs I over TCP connection.

    These attributes can be initialized in the constructor and accessed as property.

    Example:
        >>> import secsgem.secsitcp
        >>>
        >>> settings = secsgem.secsitcp.SecsITcpSettings(device_type=secsgem.common.DeviceType.EQUIPMENT)
        >>> settings.device_type
        <DeviceType.EQUIPMENT: 0>
        >>> settings.address
        '127.0.0.1'

    """

    def __init__(self, **kwargs) -> None:
        """Initialize settings."""
        super().__init__(**kwargs)

        self._connect_mode = kwargs.get("connect_mode", SecsITcpConnectMode.CLIENT)
        self._address = kwargs.get("address", "127.0.0.1")
        self._port = kwargs.get("port", 5000)

    @property
    def connect_mode(self) -> SecsITcpConnectMode:
        """Secs I over TCP connect mode.

        Default: SecsITcpConnectMode.CLIENT
        """
        return self._connect_mode

    @property
    def address(self) -> str:
        """Remote (client) or local (server) IP address.

        Default: "127.0.0.1"
        """
        return self._address

    @property
    def port(self) -> int:  # type: ignore[override]
        """TCP port of remote host.

        Default: 5000
        """
        return self._port

    def create_protocol(self) -> secsgem.common.Protocol:
        """Protocol class for this configuration."""
        from ..secsi.protocol import SecsIProtocol  # pylint: disable=import-outside-toplevel

        return SecsIProtocol(self)

    def create_connection(self) -> secsgem.common.Connection:
        """Connection class for this configuration."""
        if self.connect_mode == SecsITcpConnectMode.CLIENT:
            return secsgem.common.TcpClientConnection(self)
        return secsgem.common.TcpServerConnection(self)

    @property
    def name(self) -> str:
        """Name of this configuration."""
        return f"HSMS-{self.connect_mode}_{self.address}:{self.port}"

    def generate_thread_name(self, functionality: str) -> str:
        """Generate a unique thread name for this configuration and a provided functionality.

        Args:
            functionality: name of the functionality to generate thread name for

        Returns:
            generated thread name

        """
        return f"secsgem_SECSITCP_{functionality}_{self.connect_mode}_{self.address}:{self.port}"



"""Handler for GEM equipment."""
from __future__ import annotations

import typing

import secsgem.common
import secsgem.secs.data_items

from .alarm_capability import AlarmCapability
from .clock_capability import ClockCapability
from .collection_event_capability import CollectionEventCapability
from .control_state_machine import ControlState
from .data_value_capability import DataValueCapability
from .equipment_constants_capability import EquipmentConstantsCapability
from .handler import GemHandler
from .remote_control_capability import RemoteControlCapability
from .state_models_capability import StateModelsCapability
from .status_data_collection_capability import StatusDataCollectionCapability

if typing.TYPE_CHECKING:
    import secsgem.secs.variables


class GemEquipmentHandler(  # pylint: disable=too-many-ancestors
    AlarmCapability,
    ClockCapability,
    CollectionEventCapability,
    DataValueCapability,
    EquipmentConstantsCapability,
    RemoteControlCapability,
    StateModelsCapability,
    StatusDataCollectionCapability,
    GemHandler
):
    """Baseclass for creating equipment models. Inherit from this class and override required functions."""

    def __init__(self,
                 settings: secsgem.common.Settings,
                 initial_control_state: str = "ATTEMPT_ONLINE",
                 initial_online_control_state: str = "REMOTE"):
        """Initialize a gem equipment handler.

        Args:
            settings: communication layer settings
            initial_control_state: initial state for the control state model, one of ["EQUIPMENT_OFFLINE",
            "ATTEMPT_ONLINE", "HOST_OFFLINE", "ONLINE"]
            initial_online_control_state: initial state for online control state model

        """
        self.__initial_control_state = initial_control_state
        self.__initial_online_control_state = initial_online_control_state

        super().__init__(settings)

        self._is_host = False

    @property
    def _initial_control_state(self) -> str:
        return self.__initial_control_state

    @property
    def _initial_online_control_state(self) -> str:
        return self.__initial_online_control_state

    def on_connection_closed(self, connection):
        """Handle connection was closed event."""
        # call parent handlers
        super().on_connection_closed(connection)

        # update control state
        if self._control_state.current in [ControlState.ONLINE, ControlState.ONLINE_LOCAL, ControlState.ONLINE_REMOTE]:
            self._control_state.switch_offline()

        if self._control_state.current == ControlState.EQUIPMENT_OFFLINE:
            self._control_state.switch_online()



"""SECS array variable type."""

import secsgem.common

from .base import Base
from .list_type import List


class Array(Base):
    """List variable type. List with items of same type."""

    format_code = 0
    text_code = "L"
    preferred_types = [list]

    class _SecsVarArrayIter:
        def __init__(self, values):
            self._values = values
            self._counter = 0

        def __iter__(self):
            """Get an iterator."""
            return self

        def __next__(self):
            """Get the next item or raise StopIteration if at end of list."""
            if self._counter < len(self._values):
                i = self._counter
                self._counter += 1
                return self._values[i]

            raise StopIteration

    def __init__(self, data_format, value=None, count=-1):
        """Initialize a secs array variable.

        :param data_format: internal data definition/sample
        :type data_format: :class:`secs.variables.Base`
        :param value: initial value
        :type value: list
        :param count: number of fields in the list
        :type count: integer
        """
        super().__init__()

        self.item_decriptor = data_format
        self.count = count
        self.data = []
        if isinstance(data_format, list):
            self.name = List.get_name_from_format(data_format)
        elif hasattr(data_format, "__name__"):
            self.name = data_format.__name__
        else:
            self.name = "UNKNOWN"

        if value is not None:
            self.set(value)

    @staticmethod
    def get_format(data_format, showname=False):
        """Get the format of the variable.

        :returns: returns the string representation of the function
        :rtype: string
        """
        if showname:
            array_name = "{}: "
            if isinstance(data_format, list):
                array_name = array_name.format(List.get_name_from_format(data_format))
            else:
                array_name = array_name.format(data_format.__name__)
        else:
            array_name = ""

        if isinstance(data_format, list):
            return (
                f"{array_name}[\n"
                f"{secsgem.common.indent_block(List.get_format(data_format), 4)}\n"
                f"    ...\n]"
            )

        return (
            f"{array_name}[\n"
            f"{secsgem.common.indent_block(data_format.get_format(not showname), 4)}\n"
            f"    ...\n]"
        )

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        if len(self.data) == 0:
            return f"<{self.text_code}>"

        data = ""

        for value in self.data:
            data += f"{secsgem.common.indent_block(value.__repr__())}\n"

        return f"<{self.text_code} [{len(self.data)}]\n{data}\n>"

    def __len__(self):
        """Get the length."""
        return len(self.data)

    def __getitem__(self, key):
        """Get an item using the indexer operator."""
        return self.data[key]

    def __iter__(self):
        """Get an iterator."""
        return Array._SecsVarArrayIter(self.data)

    def __setitem__(self, key, value):
        """Set an item using the indexer operator."""
        if isinstance(value, (type(self.data[key]), self.data[key].__class__.__bases__)):
            self.data[key] = value
        elif isinstance(value, Base):
            raise TypeError(f"Wrong type {value.__class__.__name__} when expecting {self.data[key].__class__.__name__}")
        else:
            self.data[key].set(value)

    def append(self, data):
        """Append data to the internal list.

        :param value: new value
        :type value: various
        """
        from .functions import generate  # pylint: disable=import-outside-toplevel,cyclic-import

        new_object = generate(self.item_decriptor)
        new_object.set(data)
        self.data.append(new_object)

    def set(self, value):
        """Set the internal value to the provided value.

        :param value: new value
        :type value: list
        """
        if not isinstance(value, list):
            raise TypeError(f"Invalid value type {type(value).__name__} for {self.__class__.__name__}")

        if self.count >= 0 and not len(value) == self.count:
            raise ValueError(f"Value has invalid field count (expected: {self.count}, actual: {len(value)})")

        self.data = []

        for item in value:
            from .functions import generate  # pylint: disable=import-outside-toplevel,cyclic-import
            new_object = generate(self.item_decriptor)
            new_object.set(item)
            self.data.append(new_object)

    def get(self):
        """Return the internal value.

        :returns: internal value
        :rtype: list
        """
        return [item.get() for item in self.data]

    def encode(self):
        """Encode the value to secs data.

        :returns: encoded data bytes
        :rtype: string
        """
        result = self.encode_item_header(len(self.data))

        for item in self.data:
            result += item.encode()

        return result

    def decode(self, data, start=0):
        """Decode the secs byte data to the value.

        :param data: encoded data bytes
        :type data: string
        :param start: start position of value the data
        :type start: integer
        :returns: new start position
        :rtype: integer
        """
        from .functions import generate  # pylint: disable=import-outside-toplevel,cyclic-import

        (text_pos, _, length) = self.decode_item_header(data, start)

        # list
        self.data = []

        for _ in range(length):
            new_object = generate(self.item_decriptor)
            text_pos = new_object.decode(data, text_pos)
            self.data.append(new_object)

        return text_pos



"""EXID data item."""
from .. import variables
from .base import DataItemBase


class EXID(DataItemBase):
    """Exception identifier.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 20

    **Used In Function**
        - :class:`SecsS05F09 <secsgem.secs.functions.SecsS05F09>`
        - :class:`SecsS05F11 <secsgem.secs.functions.SecsS05F11>`
        - :class:`SecsS05F13 <secsgem.secs.functions.SecsS05F13>`
        - :class:`SecsS05F14 <secsgem.secs.functions.SecsS05F14>`
        - :class:`SecsS05F15 <secsgem.secs.functions.SecsS05F15>`
        - :class:`SecsS05F17 <secsgem.secs.functions.SecsS05F17>`
        - :class:`SecsS05F18 <secsgem.secs.functions.SecsS05F18>`

    """

    __type__ = variables.String
    __count__ = 20



"""Class for stream 02 function 48."""

from secsgem.secs.data_items import LIMITID, LIMITMAX, LIMITMIN, LOWERDB, UNITS, UPPERDB, VID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F48(SecsStreamFunction):
    """Define variable limit attributes - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F48
        [
            {
                VID: U1/U2/U4/U8/I1/I2/I4/I8/A
                DATA: {
                    UNITS: A
                    LIMITMIN: BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A[1]
                    LIMITMAX: BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A[1]
                    DATA: [
                        {
                            LIMITID: B[1]
                            UPPERDB: BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A[1]
                            LOWERDB: BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A[1]
                        }
                        ...
                    ]
                }
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F48([{
        ...     "VID": 1,
        ...     "DATA": {
        ...         "UNITS": "meters",
        ...         "LIMITMIN": 10,
        ...         "LIMITMAX": 20,
        ...         "DATA": [{
        ...             "LIMITID": 3,
        ...             "UPPERDB": 50,
        ...             "LOWERDB": 5
        ...         }]
        ...     }}])
        S2F48
          <L [1]
            <L [2]
              <U1 1 >
              <L [4]
                <A "meters">
                <U1 10 >
                <U1 20 >
                <L [1]
                  <L [3]
                    <B 0x3>
                    <U1 50 >
                    <U1 5 >
                  >
                >
              >
            >
          > .

    Data Items:
        - :class:`VID <secsgem.secs.data_items.VID>`
        - :class:`UNITS <secsgem.secs.data_items.UNITS>`
        - :class:`LIMITMIN <secsgem.secs.data_items.LIMITMIN>`
        - :class:`LIMITMAX <secsgem.secs.data_items.LIMITMAX>`
        - :class:`LIMITID <secsgem.secs.data_items.LIMITID>`
        - :class:`UPPERDB <secsgem.secs.data_items.UPPERDB>`
        - :class:`LOWERDB <secsgem.secs.data_items.LOWERDB>`

    """

    _stream = 2
    _function = 48

    _data_format = [
        [
            VID,
            [
                UNITS,
                LIMITMIN,
                LIMITMAX,
                [
                    [
                        LIMITID,
                        UPPERDB,
                        LOWERDB
                    ]
                ]
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Wrapper for GEM remote command."""
from __future__ import annotations

import enum
import typing

from secsgem.secs.variables import U4, Base, String

if typing.TYPE_CHECKING:
    from .collection_event import CollectionEventId


class RemoteCommandId(enum.Enum):
    """Default IDs for remote commands."""
    START = "START"
    STOP = "STOP"

class RemoteCommand:  # pylint: disable=too-few-public-methods
    """Remote command definition."""

    def __init__(self,
                 rcmd: int | str | RemoteCommandId,
                 name: str,
                 params: list[str],
                 ce_finished: int | str | CollectionEventId,
                 **kwargs):
        """Initialize a remote command.

        You can manually set the secs-type of the id with the 'id_type' keyword argument.

        Custom parameters can be set with the keyword arguments,
        they will be passed to the GemEquipmentHandlers callback
        :func:`secsgem.gem.equipmenthandler.GemEquipmentHandler._on_rcmd_<remote_command>`.

        :param rcmd: ID of the status variable
        :type rcmd: various
        :param name: long name of the status variable
        :type name: string
        :param params: array of available parameter names
        :type params: list
        :param ce_finished: collection event to trigger when remote command was finished
        :type ce_finished: types supported by data item CEID
        """
        self.rcmd = rcmd
        self.name = name
        self.params = params
        self.ce_finished = ce_finished

        self.id_type: type[Base]

        if isinstance(self.rcmd, int):
            self.id_type = U4
        else:
            self.id_type = String

        for key, value in kwargs.items():
            setattr(self, key, value)



"""MDLN data item."""
from .. import variables
from .base import DataItemBase


class MDLN(DataItemBase):
    """Equipment model type.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 20

    **Used In Function**
        - :class:`SecsS01F02 <secsgem.secs.functions.SecsS01F02>`
        - :class:`SecsS01F13 <secsgem.secs.functions.SecsS01F13>`
        - :class:`SecsS01F14 <secsgem.secs.functions.SecsS01F14>`

    """

    __type__ = variables.String
    __count__ = 20



"""Class for stream 02 function 29."""

from secsgem.secs.data_items import ECID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F29(SecsStreamFunction):
    """equipment constant namelist - request.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F29
        [
            ECID: U1/U2/U4/U8/I1/I2/I4/I8/A
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F29([1, 1337])
        S2F29 W
          <L [2]
            <U1 1 >
            <U2 1337 >
          > .

    Data Items:
        - :class:`ECID <secsgem.secs.data_items.ECID>`

    An empty list will return all available equipment constants.

    """

    _stream = 2
    _function = 29

    _data_format = [ECID]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""SECS variable helper functions."""

import inspect

from .base import Base


def generate(data_format):
    """Generate actual variable from data format.

    :param data_format: data format to create variable for
    :type data_format: list/Base based class
    :returns: created variable
    :rtype: Base based class
    """
    from .array import Array  # pylint: disable=import-outside-toplevel,cyclic-import
    from .list_type import List  # pylint: disable=import-outside-toplevel,cyclic-import

    if data_format is None:
        return None

    if isinstance(data_format, list):
        if len(data_format) == 1:
            return Array(data_format[0])
        return List(data_format)
    if inspect.isclass(data_format):
        if issubclass(data_format, Base):
            return data_format()
        raise TypeError(f"Can't generate item of class {data_format.__name__}")
    raise TypeError(f"Can't handle item of class {data_format.__class__.__name__}")


def get_format(data_format, showname=False):
    """Get the format of the function.

    :returns: returns the string representation of the function
    :rtype: string
    """
    del showname  # unused variable

    from .array import Array  # pylint: disable=import-outside-toplevel,cyclic-import
    from .list_type import List  # pylint: disable=import-outside-toplevel,cyclic-import

    if data_format is None:
        return None

    if isinstance(data_format, list):
        if len(data_format) == 1:
            return Array.get_format(data_format[0])
        return List.get_format(data_format)

    if inspect.isclass(data_format):
        if issubclass(data_format, Base):
            return data_format.get_format()
        raise TypeError(f"Can't generate data_format for class {data_format.__name__}")

    raise TypeError(f"Can't handle item of class {data_format.__class__.__name__}")



"""Alarm Management capability."""
from __future__ import annotations

import typing

import secsgem.secs

from .capability import Capability
from .handler import GemHandler

if typing.TYPE_CHECKING:
    from .alarm import Alarm


class AlarmCapability(GemHandler, Capability):
    """Alarm Management capability for GEM."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize capability."""
        super().__init__(*args, **kwargs)

        self.__alarms: dict[int | str, Alarm] = {
        }

    @property
    def _alarms(self) -> dict[int | str, Alarm]:
        return self.__alarms

    @property
    def alarms(self) -> dict[int | str, Alarm]:
        """Get the list of the alarms.

        Returns:
            Alarms list

        """
        return self._alarms

    def set_alarm(self, alid: int | str):
        """Set the list of the alarms.

        Args:
            alid: Alarm id

        """
        if alid not in self.alarms:
            raise ValueError(f"Unknown alarm id {alid}")

        if self.alarms[alid].set:
            return

        if self.alarms[alid].enabled:
            self.send_and_waitfor_response(self.stream_function(5, 1)(
                {
                    "ALCD": self.alarms[alid].code | secsgem.secs.data_items.ALCD.ALARM_SET,
                    "ALID": alid,
                    "ALTX": self.alarms[alid].text,
                }))

        self.alarms[alid].set = True

        self.trigger_collection_events([self.alarms[alid].ce_on])

    def clear_alarm(self, alid: int | str):
        """Clear the list of the alarms.

        Args:
            alid: Alarm id

        """
        if alid not in self.alarms:
            raise ValueError(f"Unknown alarm id {alid}")

        if not self.alarms[alid].set:
            return

        if self.alarms[alid].enabled:
            self.send_and_waitfor_response(self.stream_function(5, 1)({"ALCD": self.alarms[alid].code,
                                                                       "ALID": alid, "ALTX": self.alarms[alid].text}))

        self.alarms[alid].set = False

        self.trigger_collection_events([self.alarms[alid].ce_off])

    def _on_s05f03(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 5, Function 3, Alarm en-/disabled.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        result = secsgem.secs.data_items.ACKC5.ACCEPTED

        alid = function.ALID.get()
        if alid not in self._alarms:
            result = secsgem.secs.data_items.ACKC5.ERROR
        else:
            self.alarms[alid].enabled = function.ALED.get() == secsgem.secs.data_items.ALED.ENABLE

        return self.stream_function(5, 4)(result)

    def _on_s05f05(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 5, Function 5, Alarm list.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler  # unused parameters

        function = self.settings.streams_functions.decode(message)

        alids = function.get()

        if len(alids) == 0:
            alids = list(self.alarms.keys())

        result = [{
            "ALCD": self.alarms[alid].code | (secsgem.secs.data_items.ALCD.ALARM_SET if self.alarms[alid].set else 0),
            "ALID": alid,
            "ALTX": self.alarms[alid].text,
        } for alid in alids]

        return self.stream_function(5, 6)(result)

    def _on_s05f07(self,
                   handler: secsgem.secs.SecsHandler,
                   message: secsgem.common.Message) -> secsgem.secs.SecsStreamFunction | None:
        """Handle Stream 5, Function 7, Enabled alarm list.

        Args:
            handler: handler the message was received on
            message: complete message received

        """
        del handler, message  # unused parameters

        result = [{
            "ALCD": self.alarms[alid].code | (secsgem.secs.data_items.ALCD.ALARM_SET if self.alarms[alid].set else 0),
            "ALID": alid,
            "ALTX": self.alarms[alid].text,
        } for alid in list(self.alarms.keys()) if self.alarms[alid].enabled]

        return self.stream_function(5, 8)(result)

    def _get_alarms_enabled(self) -> list[int | str]:
        """List of the enabled alarms.

        :returns: alarms
        :rtype: list of various
        """
        enabled_alarms = []

        for alid, alarm in self._alarms.items():
            if alarm.enabled:
                enabled_alarms.append(alid)

        return enabled_alarms

    def _get_alarms_set(self) -> list[int | str]:
        """List of the set alarms.

        :returns: alarms
        :rtype: list of various
        """
        set_alarms = []

        for alid, alarm in self._alarms.items():
            if alarm.set:
                set_alarms.append(alid)

        return set_alarms



"""LIMITID data item."""
from .. import variables
from .base import DataItemBase


class LIMITID(DataItemBase):
    """Limit ID.

    :Type: :class:`Binary <secsgem.secs.variables.Binary>`
    :Length: 1

    **Used In Function**
        - :class:`SecsS02F45 <secsgem.secs.functions.SecsS02F45>`
        - :class:`SecsS02F46 <secsgem.secs.functions.SecsS02F46>`
        - :class:`SecsS02F48 <secsgem.secs.functions.SecsS02F48>`

    """

    __type__ = variables.Binary
    __count__ = 1



"""Class for stream 02 function 15."""

from secsgem.secs.data_items import ECID, ECV
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F15(SecsStreamFunction):
    """new equipment constant - send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F15
        [
            {
                ECID: U1/U2/U4/U8/I1/I2/I4/I8/A
                ECV: L/BOOLEAN/I8/I1/I2/I4/F8/F4/U8/U1/U2/U4/A/B
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F15([
        ...     {"ECID": 1, "ECV": secsgem.secs.variables.U4(10)},
        ...     {"ECID": "1337", "ECV": "text"}])
        S2F15 W
          <L [2]
            <L [2]
              <U1 1 >
              <U4 10 >
            >
            <L [2]
              <A "1337">
              <A "text">
            >
          > .

    Data Items:
        - :class:`ECID <secsgem.secs.data_items.ECID>`
        - :class:`ECV <secsgem.secs.data_items.ECV>`

    """

    _stream = 2
    _function = 15

    _data_format = [
        [
            ECID,
            ECV
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Handler for SECS commands."""
from __future__ import annotations

import logging
import typing

import secsgem.common
import secsgem.hsms


class SecsHandler:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Baseclass for creating Host/Equipment models. This layer contains the SECS functionality.

    Inherit from this class and override required functions.
    """

    def __init__(self, settings: secsgem.common.Settings):
        """Initialize a secs handler.

        Args:
            settings: settings defining protocol and connection

        """
        self._settings = settings

        self._protocol = settings.create_protocol()
        self._protocol.events.message_received += self._on_message_received

        self.logger = logging.getLogger(self.__module__ + "." + self.__class__.__name__)

        self._callback_handler = secsgem.common.CallbackHandler()
        self._callback_handler.target = self

    @property
    def settings(self) -> secsgem.common.Settings:
        """Get the setting object."""
        return self._settings

    @staticmethod
    def _generate_sf_callback_name(stream: int, function: int) -> str:
        return f"s{stream:02d}f{function:02d}"

    @property
    def protocol(self) -> secsgem.common.Protocol:
        """Get the connection for the handler."""
        return self._protocol

    def enable(self):
        """Enable the connection."""
        self.protocol.enable()

    def disable(self):
        """Disable the connection."""
        self.protocol.disable()

    def send_response(self, *args, **kwargs):
        """Wrapper for connections send_response function."""
        return self.protocol.send_response(*args, **kwargs)

    def send_and_waitfor_response(self, *args, **kwargs):
        """Wrapper for connections send_and_waitfor_response function."""
        return self.protocol.send_and_waitfor_response(*args, **kwargs)

    def send_stream_function(self, *args, **kwargs):
        """Wrapper for connections send_stream_function function."""
        return self.protocol.send_stream_function(*args, **kwargs)

    @property
    def events(self):
        """Wrapper for connections events."""
        return self.protocol.events

    @property
    def callbacks(self):
        """Property for callback handling."""
        return self._callback_handler

    def register_stream_function(self, stream: int, function: int, callback):
        """Register the function callback for stream and function.

        :param stream: stream to register callback for
        :type stream: integer
        :param function: function to register callback for
        :type function: integer
        :param callback: method to call when stream and functions is received
        :type callback: def callback(connection)
        """
        name = self._generate_sf_callback_name(stream, function)
        setattr(self._callback_handler, name, callback)

    def unregister_stream_function(self, stream, function):
        """Unregister the function callback for stream and function.

        :param stream: stream to unregister callback for
        :type stream: integer
        :param function: function to register callback for
        :type function: integer
        """
        name = self._generate_sf_callback_name(stream, function)
        setattr(self._callback_handler, name, None)

    def _handle_stream_function(self, message):
        sf_callback_index = self._generate_sf_callback_name(message.header.stream, message.header.function)

        # return S09F05 if no callback present
        if sf_callback_index not in self._callback_handler:
            self.logger.warning("unexpected function received %s\n%s", sf_callback_index, message.header)
            if message.header.require_response:
                self.send_response(self.stream_function(9, 5)(message.header.encode()), message.header.system)

            return

        try:
            callback = getattr(self._callback_handler, sf_callback_index)
            result = callback(self, message)
            if result is not None:
                self.send_response(result, message.header.system)
            if sf_callback_func := getattr(self, f"{sf_callback_index}_callback", None):
                sf_callback_func(message)
        except Exception:  # pylint: disable=broad-except
            self.logger.exception("Callback aborted because of exception, abort sent")
            self.send_response(self.stream_function(message.header.stream, 0)(), message.header.system)

    def _on_message_received(self, data: dict[str, typing.Any]):
        """Message received from protocol layer.

        Args:
            data: received data

        """
        message = data["message"]

        # check if callbacks available for this stream and function
        self._handle_stream_function(message)

    def disable_ceids(self):
        """Disable all Collection Events."""
        self.logger.info("Disable all collection events")

        return self.send_and_waitfor_response(self.stream_function(2, 37)({"CEED": False, "CEID": []}))

    def disable_ceid_reports(self):
        """Disable all Collection Event Reports."""
        self.logger.info("Disable all collection event reports")

        return self.send_and_waitfor_response(self.stream_function(2, 33)({"DATAID": 0, "DATA": []}))

    def list_svs(self, svs=None):
        """Get list of available Service Variables.

        :returns: available Service Variables
        :rtype: list
        """
        self.logger.info("Get list of service variables")

        if svs is None:
            svs = []

        message = self.send_and_waitfor_response(self.stream_function(1, 11)(svs))

        return self.settings.streams_functions.decode(message)

    def request_svs(self, svs):
        """Request contents of supplied Service Variables.

        :param svs: Service Variables to request
        :type svs: list
        :returns: values of requested Service Variables
        :rtype: list
        """
        self.logger.info("Get value of service variables %s", svs)

        message = self.send_and_waitfor_response(self.stream_function(1, 3)(svs))

        return self.settings.streams_functions.decode(message)

    def request_sv(self, sv_id):
        """Request contents of one Service Variable.

        :param sv_id: id of Service Variable
        :type sv_id: int
        :returns: value of requested Service Variable
        :rtype: various
        """
        self.logger.info("Get value of service variable %s", sv_id)

        return self.request_svs([sv_id])[0]

    def list_ecs(self, ecs=None):
        """Get list of available Equipment Constants.

        :returns: available Equipment Constants
        :rtype: list
        """
        self.logger.info("Get list of equipment constants")

        if ecs is None:
            ecs = []
        message = self.send_and_waitfor_response(self.stream_function(2, 29)(ecs))

        return self.settings.streams_functions.decode(message)

    def request_ecs(self, ecs):
        """Request contents of supplied Equipment Constants.

        :param ecs: Equipment Constants to request
        :type ecs: list
        :returns: values of requested Equipment Constants
        :rtype: list
        """
        self.logger.info("Get value of equipment constants %s", ecs)

        message = self.send_and_waitfor_response(self.stream_function(2, 13)(ecs))

        return self.settings.streams_functions.decode(message)

    def request_ec(self, ec_id):
        """Request contents of one Equipment Constant.

        :param ec_id: id of Equipment Constant
        :type ec_id: int
        :returns: value of requested Equipment Constant
        :rtype: various
        """
        self.logger.info("Get value of equipment constant %s", ec_id)

        return self.request_ecs([ec_id])

    def set_ecs(self, ecs):
        """Set contents of supplied Equipment Constants.

        :param ecs: list containing list of id / value pairs
        :type ecs: list
        """
        self.logger.info("Set value of equipment constants %s", ecs)

        message = self.send_and_waitfor_response(self.stream_function(2, 15)(ecs))

        return self.settings.streams_functions.decode(message).get()

    def set_ec(self, ec_id, value):
        """Set contents of one Equipment Constant.

        :param ec_id: id of Equipment Constant
        :type ec_id: int
        :param value: new content of Equipment Constant
        :type value: various
        """
        self.logger.info("Set value of equipment constant %s to %s", ec_id, value)

        return self.set_ecs([[ec_id, value]])

    def send_equipment_terminal(self, terminal_id, text):
        """Set text to equipment terminal.

        :param terminal_id: ID of terminal
        :type terminal_id: int
        :param text: text to send
        :type text: string
        """
        self.logger.info("Send text to terminal %s", terminal_id)

        return self.send_and_waitfor_response(self.stream_function(10, 3)({"TID": terminal_id, "TEXT": text}))

    def are_you_there(self):
        """Check if remote is still replying."""
        self.logger.info("Requesting 'are you there'")

        return self.send_and_waitfor_response(self.stream_function(1, 1)())

    def stream_function(self, stream: int, function: int) -> type[secsgem.secs.SecsStreamFunction]:
        """Get class for stream and function.

        Args:
            stream: stream to get class for
            function: function to get class for

        Returns:
            class for function

        """
        klass = self.settings.streams_functions.function(stream, function)

        if klass is None:
            raise KeyError(f"Undefined function requested: S{stream:02d}F{function:02d}")

        return klass



"""Class for stream 05 function 12."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F12(SecsStreamFunction):
    """exception clear - confirm.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F12
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F12()
        S5F12 .

    """

    _stream = 5
    _function = 12

    _data_format = None

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""BCEQU data item."""
from .. import variables
from .base import DataItemBase


class BCEQU(DataItemBase):
    """Bin code equivalents.

    :Types:
       - :class:`U1 <secsgem.secs.variables.U1>`
       - :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS12F03 <secsgem.secs.functions.SecsS12F03>`
        - :class:`SecsS12F04 <secsgem.secs.functions.SecsS12F04>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.U1,
        variables.String
    ]



"""Data items module initializer."""
from .base import DataItemBase

# isort: split

from .abs import ABS
from .acka import ACKA
from .ackc5 import ACKC5
from .ackc6 import ACKC6
from .ackc7 import ACKC7
from .ackc10 import ACKC10
from .alcd import ALCD
from .aled import ALED
from .alid import ALID
from .altx import ALTX
from .attrdata import ATTRDATA
from .attrid import ATTRID
from .attrreln import ATTRRELN
from .bcequ import BCEQU
from .binlt import BINLT
from .caack import CAACK
from .carrieraction import CARRIERACTION
from .carrierid import CARRIERID
from .cattrdata import CATTRDATA
from .cattrid import CATTRID
from .ceed import CEED
from .ceid import CEID
from .cename import CENAME
from .cepack import CEPACK
from .cepval import CEPVAL
from .cmda import CMDA
from .colct import COLCT
from .commack import COMMACK
from .cpack import CPACK
from .cpname import CPNAME
from .cpval import CPVAL
from .dataid import DATAID
from .datalength import DATALENGTH
from .datlc import DATLC
from .drack import DRACK
from .dsid import DSID
from .dsper import DSPER
from .dutms import DUTMS
from .dvname import DVNAME
from .dvval import DVVAL
from .dvvalname import DVVALNAME
from .eac import EAC
from .ecdef import ECDEF
from .ecid import ECID
from .ecmax import ECMAX
from .ecmin import ECMIN
from .ecname import ECNAME
from .ecv import ECV
from .edid import EDID
from .erack import ERACK
from .errcode import ERRCODE
from .errtext import ERRTEXT
from .exid import EXID
from .exmessage import EXMESSAGE
from .exrecvra import EXRECVRA
from .extype import EXTYPE
from .fcnid import FCNID
from .ffrot import FFROT
from .fnloc import FNLOC
from .grant6 import GRANT6
from .grnt1 import GRNT1
from .hcack import HCACK
from .idtyp import IDTYP
from .length import LENGTH
from .limitack import LIMITACK
from .limitid import LIMITID
from .limitmax import LIMITMAX
from .limitmin import LIMITMIN
from .lowerdb import LOWERDB
from .lrack import LRACK
from .lvack import LVACK
from .maper import MAPER
from .mapft import MAPFT
from .mdack import MDACK
from .mdln import MDLN
from .mexp import MEXP
from .mhead import MHEAD
from .mid import MID
from .mlcl import MLCL
from .nulbc import NULBC
from .objack import OBJACK
from .objid import OBJID
from .objspec import OBJSPEC
from .objtype import OBJTYPE
from .oflack import OFLACK
from .onlack import ONLACK
from .orloc import ORLOC
from .ppbody import PPBODY
from .ppgnt import PPGNT
from .ppid import PPID
from .praxi import PRAXI
from .prdct import PRDCT
from .ptn import PTN
from .rcmd import RCMD
from .refp import REFP
from .repgsz import REPGSZ
from .rowct import ROWCT
from .rpsel import RPSEL
from .rptid import RPTID
from .rsda import RSDA
from .rsdc import RSDC
from .rsinf import RSINF
from .rspack import RSPACK
from .sdack import SDACK
from .sdbin import SDBIN
from .shead import SHEAD
from .smpln import SMPLN
from .softrev import SOFTREV
from .stime import STIME
from .strack import STRACK
from .strid import STRID
from .strp import STRP
from .sv import SV
from .svid import SVID
from .svname import SVNAME
from .text import TEXT
from .tiaack import TIAACK
from .tid import TID
from .time import TIME
from .timestamp import TIMESTAMP
from .totsmp import TOTSMP
from .trid import TRID
from .units import UNITS
from .upperdb import UPPERDB
from .v import V
from .vid import VID
from .vlaack import VLAACK
from .xdies import XDIES
from .xypos import XYPOS
from .ydies import YDIES
from .ccode import CCODE
from .pparm import PPARM

__all__ = [
    "DataItemBase",
    "ABS",
    "ACKA",
    "ACKC5",
    "ACKC6",
    "ACKC7",
    "ACKC10",
    "ALCD",
    "ALED",
    "ALID",
    "ALTX",
    "ATTRDATA",
    "ATTRID",
    "ATTRRELN",
    "BCEQU",
    "BINLT",
    "CAACK",
    "CARRIERACTION",
    "CARRIERID",
    "CATTRDATA",
    "CATTRID",
    "CEED",
    "CEID",
    "CENAME",
    "CEPACK",
    "CEPVAL",
    "CMDA",
    "COLCT",
    "COMMACK",
    "CPACK",
    "CPNAME",
    "CPVAL",
    "DATAID",
    "DATALENGTH",
    "DATLC",
    "DRACK",
    "DSID",
    "DSPER",
    "DUTMS",
    "DVNAME",
    "DVVAL",
    "DVVALNAME",
    "EAC",
    "ECDEF",
    "ECID",
    "ECMAX",
    "ECMIN",
    "ECNAME",
    "ECV",
    "EDID",
    "ERACK",
    "ERRCODE",
    "ERRTEXT",
    "EXID",
    "EXMESSAGE",
    "EXRECVRA",
    "EXTYPE",
    "FCNID",
    "FFROT",
    "FNLOC",
    "GRANT6",
    "GRNT1",
    "HCACK",
    "IDTYP",
    "LENGTH",
    "LIMITACK",
    "LIMITID",
    "LIMITMAX",
    "LIMITMIN",
    "LOWERDB",
    "LRACK",
    "LVACK",
    "MAPER",
    "MAPFT",
    "MDACK",
    "MDLN",
    "MEXP",
    "MHEAD",
    "MID",
    "MLCL",
    "NULBC",
    "OBJACK",
    "OBJID",
    "OBJSPEC",
    "OBJTYPE",
    "OFLACK",
    "ONLACK",
    "ORLOC",
    "PPBODY",
    "PPGNT",
    "PPID",
    "PRAXI",
    "PRDCT",
    "PTN",
    "RCMD",
    "REFP",
    "REPGSZ",
    "ROWCT",
    "RPSEL",
    "RPTID",
    "RSDA",
    "RSDC",
    "RSINF",
    "RSPACK",
    "SDACK",
    "SDBIN",
    "SHEAD",
    "SMPLN",
    "SOFTREV",
    "STIME",
    "STRACK",
    "STRID",
    "STRP",
    "SV",
    "SVID",
    "SVNAME",
    "TEXT",
    "TIAACK",
    "TID",
    "TIME",
    "TIMESTAMP",
    "TOTSMP",
    "TRID",
    "UNITS",
    "UPPERDB",
    "V",
    "VID",
    "VLAACK",
    "XDIES",
    "XYPOS",
    "YDIES",
    "CCODE",
    "PPARM"
]



"""Class for stream 01 function 22."""

from secsgem.secs.data_items import DVVALNAME, UNITS, VID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS01F22(SecsStreamFunction):
    """Data variable namelist.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F22
        [
            {
                VID: U1/U2/U4/U8/I1/I2/I4/I8/A
                DVVALNAME: A
                UNITS: A
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS01F22([{"VID": 1, "DVVALNAME": "EC1", "UNITS": "mm"}])
        S1F22
          <L [1]
            <L [3]
              <U1 1 >
              <A "EC1">
              <A "mm">
            >
          > .

    Data Items:
        - :class:`VID <secsgem.secs.data_items.VID>`
        - :class:`DVVALNAME <secsgem.secs.data_items.DVVALNAME>`
        - :class:`UNITS <secsgem.secs.data_items.UNITS>`

    """

    _stream = 1
    _function = 22

    _data_format = [
        [
            VID,
            DVVALNAME,
            UNITS
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""TCP client connection."""
from __future__ import annotations

import socket
import threading
import time
import typing

from .tcp_connection import TcpConnection

if typing.TYPE_CHECKING:
    import secsgem.common


class TcpClientConnection(TcpConnection):
    """Client class for single tcp client connection."""

    def __init__(self, settings: secsgem.common.Settings):
        """Initialize a TCP client connection.

        Args:
            settings: protocol and communication settings

        """
        # initialize super class
        TcpConnection.__init__(self, settings)

        # initially not enabled
        self.enabled = False

        # reconnect thread required for client connection
        self.connection_thread = None
        self.stop_connection_thread = False

        # flag if this is the first connection since enable
        self.first_connection = True

        self.on_disconnected.register(self._disconnected)

    def _disconnected(self, _: dict[str, typing.Any]):
        """Signal from super that the connection was closed.

        This is required to initiate the reconnect if the connection is still enabled
        """
        if self.enabled:
            self.__start_connect_thread()

    def enable(self):
        """Enable the connection.

        Starts the client connection process to the remote.
        """
        # only start if not already enabled
        if not self.enabled:
            # reset first connection to eliminate reconnection timeout
            self.first_connection = True

            # mark connection as enabled
            self.enabled = True

            # start the connection thread
            self.__start_connect_thread()

    def disable(self):
        """Disable the connection.

        Stops all connection attempts, and closes the connection
        """
        # only stop if enabled
        if self.enabled:
            # mark connection as disabled
            self.enabled = False

            # stop connection thread if it is running
            if self.connection_thread and self.connection_thread.is_alive():
                self.stop_connection_thread = True

            # wait for connection thread to stop
            while self.stop_connection_thread:
                time.sleep(0.2)

            # disconnect super class
            self.disconnect()

    def __idle(self, timeout: float):
        """Wait until timeout elapsed or connection thread is stopped.

        Args:
            timeout: number of seconds to wait

        Returns:
                False if thread was stopped

        """
        for _ in range(int(timeout) * 5):
            time.sleep(0.2)

            # check if connection was disabled
            if self.stop_connection_thread:
                self.stop_connection_thread = False
                return False

        return True

    def __start_connect_thread(self):
        self.connection_thread = threading.Thread(
            target=self.__connect_thread,
            name=f"secsgem_tcpClientConnection_connectThread_{self._settings.address}")
        self.connection_thread.start()

    def __connect_thread(self):
        """Thread function to (re)connect client connection to remote host.

        .. warning:: Do not call this directly, for internal use only.
        """
        # wait for timeout if this is not the first connection
        if not self.first_connection and not self.__idle(self._settings.timeouts.t5):
            return

        self.first_connection = False

        # try to connect to remote
        while not self.__connect():
            if not self.__idle(self._settings.timeouts.t5):
                return

    def __connect(self):
        """Open connection to remote host.

        Returns:
            True if connection was established, False if connection failed

        """
        # create socket
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # setup socket
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)

        self._logger.debug("connecting to %s:%d", self._settings.address, self._settings.port)

        # try to connect socket
        try:
            self._socket.connect((self._settings.address, self._settings.port))
        except OSError:
            self._logger.debug("connecting to %s:%d failed", self._settings.address, self._settings.port)
            return False

        # make socket nonblocking
        self._socket.setblocking(0)

        # mark connection as connected
        self._connected = True

        # start the receiver thread
        self._start_receiver()

        # send event
        try:
            self.on_connected({"source": self})
        except Exception:  # pylint: disable=broad-except
            self._logger.exception("ignoring exception for on_connected handler")

        return True



"""CEED data item."""
from .. import variables
from .base import DataItemBase


class CEED(DataItemBase):
    """Collection event or trace enable/disable code.

    :Type: :class:`Boolean <secsgem.secs.variables.Boolean>`
    :Length: 1

    **Values**
        +-------+---------+
        | Value |         |
        +=======+=========+
        | True  | Enable  |
        +-------+---------+
        | False | Disable |
        +-------+---------+

    **Used In Function**
        - :class:`SecsS02F37 <secsgem.secs.functions.SecsS02F37>`

    """

    __type__ = variables.Boolean
    __count__ = 1



"""Contains class to create model for hsms endpoints."""
from __future__ import annotations

import queue
import struct
import threading
import typing

import secsgem.common

from .connection_state_machine import ConnectionState, ConnectionStateMachine
from .deselect_req_header import HsmsDeselectReqHeader
from .deselect_rsp_header import HsmsDeselectRspHeader
from .header import HsmsSType
from .linktest_req_header import HsmsLinktestReqHeader
from .linktest_rsp_header import HsmsLinktestRspHeader
from .message import HsmsBlock, HsmsMessage
from .reject_req_header import HsmsRejectReqHeader
from .select_req_header import HsmsSelectReqHeader
from .select_rsp_header import HsmsSelectRspHeader
from .separate_req_header import HsmsSeparateReqHeader
from .stream_function_header import HsmsStreamFunctionHeader

if typing.TYPE_CHECKING:
    from ..secs.functions.base import SecsStreamFunction
    from .settings import HsmsSettings


class HsmsProtocol(secsgem.common.Protocol[HsmsMessage, HsmsBlock]):  # pylint: disable=too-many-instance-attributes
    """Baseclass for creating Host/Equipment models.

    This layer contains the HSMS functionality.
    Inherit from this class and override required functions.
    """

    send_packet_size = 1024 * 1024
    """ Block size for outbound data ."""

    message_type = HsmsMessage

    def __init__(self, settings: HsmsSettings):
        """Initialize hsms handler.

        Args:
            settings: protocol and communication settings

        Example:
            import secsgem.hsms

            settings = secsgem.hsms.HsmsSettings(
                address="10.211.55.33",
                port=5000,
                active=secsgem.hsms.HsmsConnectMode.ACTIVE,
                name="test"
            )
            def onConnect(event, data):
                print ("Connected")

            client = secsgem.hsms.HsmsProtocol(settings)
            client.events.connected += onConnect

            client.enable()

            time.sleep(3)

            client.disable()

        """
        super().__init__(settings)
        self._settings: HsmsSettings = settings

        self._connected = False

        # repeating linktest variables
        self._linktest_timer = None
        self._linktest_timeout = 30

        # select request thread for active connections, to avoid blocking state changes
        self._select_req_thread: threading.Thread | None = None

        # hsms connection state fsm
        self._connection_state = ConnectionStateMachine()

        self._connection_state.connected.events.enter.register(self._on_state_connect)
        self._connection_state.connected.events.leave.register(self._on_state_disconnect)
        self._connection_state.connected_selected.events.enter.register(self._on_state_select)

    @property
    def connection_state(self) -> ConnectionStateMachine:
        """Property for connection state."""
        return self._connection_state

    def _send_select_req_thread(self):
        try:
            response = self.send_select_req()
            if response is None:
                self._logger.warning("select request failed")
        except Exception as exc:  # pylint: disable=broad-exception-caught
            self._logger.warning("exception in _send_select_req_thread", exc_info=exc)

    def _start_linktest_timer(self):
        self._linktest_timer = threading.Timer(self._linktest_timeout, self._on_linktest_timer)
        self._linktest_timer.daemon = True  # kill thread automatically on main program termination
        self._linktest_timer.name = "secsgem_hsmsProtocol_linktestTimer"
        self._linktest_timer.start()

    def _on_state_connect(self, _: dict[str, typing.Any]):
        """Handle connection state model got event connect.

        :param data: event attributes
        :type data: object
        """
        # start linktest timer
        self._start_linktest_timer()

        # start select process if connection is active
        if self._settings.is_active:
            self._select_req_thread = threading.Thread(
                target=self._send_select_req_thread,
                name="secsgem_hsmsProtocol_sendSelectReqThread")
            self._select_req_thread.daemon = True  # kill thread automatically on main program termination
            self._select_req_thread.start()

    def _on_state_disconnect(self, _: dict[str, typing.Any]):
        """Handle connection state model got event disconnect.

        :param data: event attributes
        :type data: object
        """
        # stop linktest timer
        if self._linktest_timer:
            self._linktest_timer.cancel()

        self._linktest_timer = None

    def _on_state_select(self, _: dict[str, typing.Any]):
        """Handle connection state model got event select.

        :param data: event attributes
        :type data: object
        """
        # send event
        self.events.fire("communicating", {"connection": self})

    def _on_linktest_timer(self):
        """Linktest time timed out, so send linktest request."""
        # send linktest request and wait for response
        self.send_linktest_req()

        # restart the timer
        self._start_linktest_timer()

    def _on_connected(self, _: dict[str, typing.Any]):
        """Handle connection was established event."""
        self._connected = True

        self._thread.start()

        # update connection state
        self._connection_state.connect()

        self.events.fire("connected", {"connection": self})

    def _on_disconnecting(self, _: dict[str, typing.Any]):
        """Handle connection is about to be closed event."""
        # send separate request
        self.send_separate_req()

    def _on_disconnected(self, _: dict[str, typing.Any]):
        """Handle connection was _ event."""
        # update connection state
        self._connected = False
        self._connection_state.disconnect()

        self._thread.stop()
        self._receive_buffer.clear()

        self.events.fire("disconnected", {"connection": self})


    def __handle_hsms_requests_select_req(self, message: HsmsMessage):
        if self._connection.disconnecting:
            self.send_reject_rsp(message.header.system, message.header.s_type, 4)
        else:
            self.send_select_rsp(message.header.system)

            self._connection_state.select()

    def __handle_hsms_requests_select_rsp(self, message: HsmsMessage):
        self._connection_state.select()

        if message.header.system in self._response_queues:
            self._response_queues[message.header.system].put_nowait(message)

    def __handle_hsms_requests_deselect_req(self, message: HsmsMessage):
        if self._connection.disconnecting:
            self.send_reject_rsp(message.header.system, message.header.s_type, 4)
        else:
            self.send_deselect_rsp(message.header.system)

            self._connection_state.deselect()

    def __handle_hsms_requests_deselect_rsp(self, message: HsmsMessage):
        self._connection_state.deselect()

        if message.header.system in self._response_queues:
            self._response_queues[message.header.system].put_nowait(message)

    def __handle_hsms_requests_linktest_req(self, message: HsmsMessage):
        if self._connection.disconnecting:
            self.send_reject_rsp(message.header.system, message.header.s_type, 4)
        else:
            self.send_linktest_rsp(message.header.system)

    def __handle_hsms_requests(self, message: HsmsMessage):
        self._communication_logger.debug("< %s\n  %s", message, message.header.s_type.text, extra=self._get_log_extra())

        if message.header.s_type == HsmsSType.SELECT_REQ:
            self.__handle_hsms_requests_select_req(message)
        elif message.header.s_type == HsmsSType.SELECT_RSP:
            self.__handle_hsms_requests_select_rsp(message)
        elif message.header.s_type == HsmsSType.DESELECT_REQ:
            self.__handle_hsms_requests_deselect_req(message)
        elif message.header.s_type == HsmsSType.DESELECT_RSP:
            self.__handle_hsms_requests_deselect_rsp(message)
        elif message.header.s_type == HsmsSType.LINKTEST_REQ:
            self.__handle_hsms_requests_linktest_req(message)
        else:
            if message.header.system in self._response_queues:
                self._response_queues[message.header.system].put_nowait(message)

    def _process_received_data(self):
        if len(self._receive_buffer) < 4:
            return

        while len(self._receive_buffer) > 3:
            length_data = self._receive_buffer.wait_for(4, peek=True)
            length = struct.unpack(">L", length_data)[0] + 4

            data = self._receive_buffer.wait_for(length)

            # decode received message
            response = HsmsBlock.decode(data)

            self._thread.queue_block(self, response)

    def _on_connection_message_received(self, _: object, message: HsmsMessage):
        """Message received by connection.

        Args:
            source: source of event
            message: received data message

        """
        if message.header.s_type.value > 0:
            self.__handle_hsms_requests(message)
        else:
            decoded_message = self._settings.streams_functions.decode(message)
            self._communication_logger.info("< %s\n%s", message, decoded_message, extra=self._get_log_extra())

            if self._connection_state.current != ConnectionState.CONNECTED_SELECTED:
                self._logger.warning("received message when not selected")

                out_message = HsmsMessage(HsmsRejectReqHeader(message.header.system, message.header.s_type, 4), b"")
                self._communication_logger.info(
                    "> %s\n  %s", out_message, out_message.header.s_type.text,
                    extra=self._get_log_extra())
                self.send_message(out_message)

                return

            # someone is waiting for this message
            if message.header.system in self._response_queues:
                # send message to request sender
                self._response_queues[message.header.system].put_nowait(message)
            # just log if nobody is interested
            else:
                self.events.fire("message_received", {"connection": self, "message": message})

    def serialize_data(self) -> dict[str, typing.Any]:
        """Return data for serialization.

        :returns: data to serialize for this object
        :rtype: dict
        """
        return {
            "address": self._settings.address,
            "port": self._settings.port,
            "connect_mode": self._settings.connect_mode,
            "session_id": self._settings.session_id,
            "name": self._settings.name,
            "connected": self._connected,
        }

    def _process_send_queue(self):
        if self._send_queue.empty():
            return

        while not self._send_queue.empty():
            block_info = self._send_queue.get()

            packets = [
                block_info.data[i: i + self.send_packet_size]
                for i in range(0, len(block_info.data), self.send_packet_size)
            ]

            for packet in packets:
                if not self._connection.send_data(packet):
                    block_info.resolve(False)
                    return

            block_info.resolve(True)

    def _create_message_for_function(
            self,
            function: SecsStreamFunction,
            system_id: int,
    ) -> secsgem.common.Message:
        """Create a protocol specific message for a function.

        Args:
            function: function to create message for
            system_id: system

        Returns:
            created message

        """
        return HsmsMessage(
            HsmsStreamFunctionHeader(
                system_id,
                function.stream,
                function.function,
                function.is_reply_required,
                self._settings.session_id,
            ),
            function.encode())

    def send_select_req(self):
        """Send a Select Request to the remote host.

        :returns: System of the sent request
        :rtype: integer
        """
        system_id = self.get_next_system_counter()

        response_queue = self._get_queue_for_system(system_id)

        message = HsmsMessage(HsmsSelectReqHeader(system_id), b"")
        self._communication_logger.info(
            "> %s\n  %s", message, message.header.s_type.text,
            extra=self._get_log_extra())

        if not self.send_message(message):
            self._remove_queue(system_id)
            return None

        try:
            response = response_queue.get(True, self._settings.timeouts.t6)
        except queue.Empty:
            response = None

        self._remove_queue(system_id)

        return response

    def send_select_rsp(self, system_id):
        """Send a Select Response to the remote host.

        :param system_id: System of the request to reply for
        :type system_id: integer
        """
        message = HsmsMessage(HsmsSelectRspHeader(system_id), b"")
        self._communication_logger.info(
            "> %s\n  %s", message, message.header.s_type.text,
            extra=self._get_log_extra())
        return self.send_message(message)

    def send_linktest_req(self):
        """Send a Linktest Request to the remote host.

        :returns: System of the sent request
        :rtype: integer
        """
        system_id = self.get_next_system_counter()

        response_queue = self._get_queue_for_system(system_id)

        message = HsmsMessage(HsmsLinktestReqHeader(system_id), b"")
        self._communication_logger.debug(
            "> %s\n  %s", message, message.header.s_type.text,
            extra=self._get_log_extra())

        if not self.send_message(message):
            self._remove_queue(system_id)
            return None

        try:
            response = response_queue.get(True, self._settings.timeouts.t6)
        except queue.Empty:
            response = None

        self._remove_queue(system_id)

        return response

    def send_linktest_rsp(self, system_id):
        """Send a Linktest Response to the remote host.

        :param system_id: System of the request to reply for
        :type system_id: integer
        """
        message = HsmsMessage(HsmsLinktestRspHeader(system_id), b"")
        self._communication_logger.debug(
            "> %s\n  %s", message, message.header.s_type.text,
            extra=self._get_log_extra())
        return self.send_message(message)

    def send_deselect_req(self):
        """Send a Deselect Request to the remote host.

        :returns: System of the sent request
        :rtype: integer
        """
        system_id = self.get_next_system_counter()

        response_queue = self._get_queue_for_system(system_id)

        message = HsmsMessage(HsmsDeselectReqHeader(system_id), b"")
        self._communication_logger.info("> %s\n  %s", message, message.header.s_type.text,
                                        extra=self._get_log_extra())

        if not self.send_message(message):
            self._remove_queue(system_id)
            return None

        try:
            response = response_queue.get(True, self._settings.timeouts.t6)
        except queue.Empty:
            response = None

        self._remove_queue(system_id)

        return response

    def send_deselect_rsp(self, system_id):
        """Send a Deselect Response to the remote host.

        :param system_id: System of the request to reply for
        :type system_id: integer
        """
        message = HsmsMessage(HsmsDeselectRspHeader(system_id), b"")
        self._communication_logger.info(
            "> %s\n  %s", message, message.header.s_type.text,
            extra=self._get_log_extra())
        return self.send_message(message)

    def send_reject_rsp(self, system_id: int, s_type: HsmsSType, reason: int):
        """Send a Reject Response to the remote host.

        :param system_id: System of the request to reply for
        :type system_id: integer
        :param s_type: s_type of rejected message
        :type s_type: integer
        :param reason: reason for rejection
        :type reason: integer
        """
        message = HsmsMessage(HsmsRejectReqHeader(system_id, s_type, reason), b"")
        self._communication_logger.info(
            "> %s\n  %s", message, message.header.s_type.text,
            extra=self._get_log_extra())
        return self.send_message(message)

    def send_separate_req(self):
        """Send a Separate Request to the remote host."""
        system_id = self.get_next_system_counter()

        message = HsmsMessage(HsmsSeparateReqHeader(system_id), b"")
        self._communication_logger.info(
            "> %s\n  %s", message, message.header.s_type.text,
            extra=self._get_log_extra())

        if not self.send_message(message):
            return None

        return system_id

    # helpers

    def _get_log_extra(self) -> dict[str, typing.Any]:
        """Get extra fields for logging."""
        return {"address": self._settings.address,
                "port": self._settings.port,
                "session_id": self._settings.session_id,
                "remoteName": self._settings.name}



"""SECS string variable base type."""

import unicodedata

from .base import Base


class BaseText(Base):
    """Secs type base for any text data."""

    format_code = -1
    text_code = ""
    control_chars = "".join(chr(ch) for ch in range(256) if unicodedata.category(chr(ch))[0] == "C")
    coding = ""

    def __init__(self, value="", count=-1):
        """Initialize a secs text variable.

        :param value: initial value
        :type value: string
        :param count: number of items this value
        :type count: integer
        """
        super().__init__()

        self.value = ""
        self.count = count

        if value is not None:
            self.set(value)

    def __repr__(self):
        """Generate textual representation for an object of this class."""
        if len(self.value) == 0:
            return f"<{self.text_code}>"

        data = ""
        last_char_printable = False

        for char in self.value:
            output = char

            if char not in self.control_chars:
                if last_char_printable:
                    data += output
                else:
                    data += ' "' + output
                last_char_printable = True
            else:
                if last_char_printable:
                    data += '" ' + hex(ord(output))
                else:
                    data += " " + hex(ord(output))
                last_char_printable = False

        if last_char_printable:
            data += '"'

        return f"<{self.text_code}{data}>"

    def __len__(self):
        """Get the length."""
        return len(self.value)

    def __eq__(self, other):
        """Check equality with other object."""
        if isinstance(other, Base):
            if other.is_dynamic:
                return other.value.value == self.value
            return other.value == self.value

        return other == self.value

    def __hash__(self):
        """Get data item for hashing."""
        return hash(self.value)

    @staticmethod
    def _check_single_item_support(value):
        if isinstance(value, bool):
            return True

        if isinstance(value, int):
            if 0 <= value <= 255:
                return True
            return False

        return False

    def supports_value(self, value) -> bool:
        """Check if the current instance supports the provided value.

        :param value: value to test
        :type value: any
        """
        if isinstance(value, (list, tuple, bytearray)):
            return self._supports_value_list(value)

        if isinstance(value, bytes):
            return self._supports_value_bytes(value)

        if isinstance(value, (int, float, complex)):
            return self._supports_value_number(value)

        if isinstance(value, str):
            return self._supports_value_str(value)

        return False

    def _supports_value_list(self, value) -> bool:
        if self.count > 0 and len(value) > self.count:
            return False

        return all(self._check_single_item_support(item) for item in value)

    def _supports_value_bytes(self, value) -> bool:
        if 0 < self.count < len(value):
            return False

        return True

    def _supports_value_number(self, value) -> bool:
        if 0 < self.count < len(str(value)):
            return False
        return True

    def _supports_value_str(self, value) -> bool:
        if 0 < self.count < len(value):
            return False
        try:
            value.encode(self.coding)
        except UnicodeEncodeError:
            return False

        return True

    def set(self, value):
        """Set the internal value to the provided value.

        :param value: new value
        :type value: string/integer
        """
        if value is None:
            raise ValueError(f"{self.__class__.__name__} can't be None")

        if isinstance(value, bytes):
            value = value.decode(self.coding)
        elif isinstance(value, bytearray):
            value = bytes(value).decode(self.coding)
        elif isinstance(value, (list, tuple)):
            value = str(bytes(bytearray(value)).decode(self.coding))
        elif isinstance(value, (int, float, complex)):
            value = str(value)
        elif isinstance(value, str):
            value.encode(self.coding)  # try if it can be encoded as ascii (values 0-127)
        else:
            raise TypeError(f"Unsupported type {type(value).__name__} for {self.__class__.__name__}")

        if 0 < self.count < len(value):
            raise ValueError(f"Value longer than {self.count} chars ({len(value)} chars)")

        self.value = str(value)

    def get(self):
        """Return the internal value.

        :returns: internal value
        :rtype: string
        """
        return self.value

    def encode(self):
        """Encode the value to secs data.

        :returns: encoded data bytes
        :rtype: string
        """
        result = self.encode_item_header(len(self.value))

        result += self.value.encode(self.coding)

        return result

    def decode(self, data, start=0):
        """Decode the secs byte data to the value.

        :param data: encoded data bytes
        :type data: string
        :param start: start position of value the data
        :type start: integer
        :returns: new start position
        :rtype: integer
        """
        (text_pos, _, length) = self.decode_item_header(data, start)

        # string
        result = ""

        if length > 0:
            result = data[text_pos:text_pos + length].decode(self.coding)

        self.set(result)

        return text_pos + length



"""Class for stream 02 function 14."""

from secsgem.secs.data_items import ECV
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F14(SecsStreamFunction):
    """equipment constant - data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F14
        [
            ECV: L/BOOLEAN/I8/I1/I2/I4/F8/F4/U8/U1/U2/U4/A/B
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F14([secsgem.secs.variables.U1(1), "text"])
        S2F14
          <L [2]
            <U1 1 >
            <A "text">
          > .

    Data Items:
        - :class:`ECV <secsgem.secs.data_items.ECV>`

    """

    _stream = 2
    _function = 14

    _data_format = [ECV]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Class for stream 02 function 17."""

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F17(SecsStreamFunction):
    """date and time - request.

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F17
        Header only

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F17()
        S2F17 W .

    """

    _stream = 2
    _function = 17

    _data_format = None

    _to_host = True
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""ECNAME data item."""
from .. import variables
from .base import DataItemBase


class ECNAME(DataItemBase):
    """Equipment constant name.

    :Type: :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS02F30 <secsgem.secs.functions.SecsS02F30>`

    """

    __type__ = variables.String



"""ACKA data item."""
from .. import variables
from .base import DataItemBase


class ACKA(DataItemBase):
    """Request success.

    :Type: :class:`Boolean <secsgem.secs.variables.Boolean>`
    :Length: 1

    **Values**
        +-------+---------+
        | Value |         |
        +=======+=========+
        | True  | Success |
        +-------+---------+
        | False | Failed  |
        +-------+---------+

    **Used In Function**
        - :class:`SecsS05F14 <secsgem.secs.functions.SecsS05F14>`
        - :class:`SecsS05F15 <secsgem.secs.functions.SecsS05F15>`
        - :class:`SecsS05F18 <secsgem.secs.functions.SecsS05F18>`

    """

    __type__ = variables.Boolean
    __count__ = 1



"""RPTID data item."""
from .. import variables
from .base import DataItemBase


class RPTID(DataItemBase):
    """Report ID.

    :Types:
       - :class:`U1 <secsgem.secs.variables.U1>`
       - :class:`U2 <secsgem.secs.variables.U2>`
       - :class:`U4 <secsgem.secs.variables.U4>`
       - :class:`U8 <secsgem.secs.variables.U8>`
       - :class:`I1 <secsgem.secs.variables.I1>`
       - :class:`I2 <secsgem.secs.variables.I2>`
       - :class:`I4 <secsgem.secs.variables.I4>`
       - :class:`I8 <secsgem.secs.variables.I8>`
       - :class:`String <secsgem.secs.variables.String>`

    **Used In Function**
        - :class:`SecsS02F33 <secsgem.secs.functions.SecsS02F33>`
        - :class:`SecsS02F35 <secsgem.secs.functions.SecsS02F35>`
        - :class:`SecsS06F11 <secsgem.secs.functions.SecsS06F11>`
        - :class:`SecsS06F16 <secsgem.secs.functions.SecsS06F16>`
        - :class:`SecsS06F19 <secsgem.secs.functions.SecsS06F19>`
        - :class:`SecsS06F21 <secsgem.secs.functions.SecsS06F21>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.U1,
        variables.U2,
        variables.U4,
        variables.U8,
        variables.I1,
        variables.I2,
        variables.I4,
        variables.I8,
        variables.String
    ]



"""Class for stream 03 function 17."""

from secsgem.secs.data_items import DATAID, CARRIERACTION, CARRIERID, CATTRDATA, CATTRID, PTN

from secsgem.secs.functions.base import SecsStreamFunction


class SecsS03F17(SecsStreamFunction):
    """Carrier Action Request."""

    _stream = 3
    _function = 17

    _data_format = [
        DATAID,
        CARRIERACTION,
        CARRIERID,
        PTN,
        [
            [
                "PARAMS",
                CATTRID,
                CATTRDATA
            ]
        ]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = True



"""ALTX data item."""
from .. import variables
from .base import DataItemBase


class ALTX(DataItemBase):
    """Alarm text.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 120

    **Used In Function**
        - :class:`SecsS05F01 <secsgem.secs.functions.SecsS05F01>`
        - :class:`SecsS05F06 <secsgem.secs.functions.SecsS05F06>`
        - :class:`SecsS05F08 <secsgem.secs.functions.SecsS05F08>`

    """

    __type__ = variables.String
    __count__ = 520



"""Class for stream 02 function 36."""

from secsgem.secs.data_items import LRACK
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F36(SecsStreamFunction):
    """link event report - acknowledge.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F36
        LRACK: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F36(secsgem.secs.data_items.LRACK.CEID_UNKNOWN)
        S2F36
          <B 0x4> .

    Data Items:
        - :class:`LRACK <secsgem.secs.data_items.LRACK>`

    """

    _stream = 2
    _function = 36

    _data_format = LRACK

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""SECS 2 byte signed integer variable type."""

from .base_number import BaseNumber


class I2(BaseNumber):
    """Secs type for 2 byte signed data.

    :param value: initial value
    :type value: list/integer
    :param count: number of items this value
    :type count: integer
    """

    format_code = 0o32
    text_code = "I2"
    _base_type = int
    _min = -32768
    _max = 32767
    _bytes = 2
    _struct_code = "h"
    preferred_types = [int]



"""Class for stream 06 function 06."""

from secsgem.secs.data_items import GRANT6
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F06(SecsStreamFunction):
    """multi block data grant.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F06
        GRANT6: B[1]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F06(secsgem.secs.data_items.GRANT6.BUSY)
        S6F6
          <B 0x1> .

    Data Items:
        - :class:`GRANT6 <secsgem.secs.data_items.GRANT6>`

    """

    _stream = 6
    _function = 6

    _data_format = GRANT6

    _to_host = False
    _to_equipment = True

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 06 function 01."""

from secsgem.secs.data_items import SMPLN, STIME, SV, TRID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F01(SecsStreamFunction):
    """Trace data send.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F01
        {
            TRID: I1/I2/I4/I8/U1/U2/U4/U8/A
            SMPLN: I1/I2/I4/I8/U1/U2/U4/U8
            STIME: A
            SV: [
                DATA: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F01({
        ...     "TRID": 1,
        ...     "SMPLN": 3,
        ...     "STIME": "TIME",
        ...     "SV": [1, 4]})
        S6F1
          <L [4]
            <I1 1 >
            <I1 3 >
            <A "TIME">
            <L [2]
              <U1 1 >
              <U1 4 >
            >
          > .

    Data Items:
        - :class:`TRID <secsgem.secs.data_items.TRID>`
        - :class:`SMPLN <secsgem.secs.data_items.SMPLN>`
        - :class:`STIME <secsgem.secs.data_items.STIME>`
        - :class:`SV <secsgem.secs.data_items.SV>`

    """

    _stream = 6
    _function = 1

    _data_format = [
        TRID,
        SMPLN,
        STIME,
        [SV]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = True



"""Helper to track block send result across threads."""
from __future__ import annotations

import enum
import threading


class BlockSendResult(enum.Enum):
    """Enum for send result including not send state."""

    NOT_SENT = 0
    SENT_OK = 1
    SENT_ERROR = 2


class BlockSendInfo:
    """Container for sending block and waiting for result."""

    def __init__(self, data: bytes):
        """Initialize block send info object.

        Args:
            data: data to send.

        """
        self._data = data

        self._result = BlockSendResult.NOT_SENT
        self._result_trigger = threading.Event()

    @property
    def data(self) -> bytes:
        """Get the data for sending."""
        return self._data

    def resolve(self, result: bool):
        """Resolve the send data with a result.

        Args:
            result: result to resolve with

        """
        self._result = BlockSendResult.SENT_OK if result else BlockSendResult.SENT_ERROR
        self._result_trigger.set()

    def wait(self) -> bool:
        """Wait for the message is sent and a result is available."""
        self._result_trigger.wait()

        return self._result == BlockSendResult.SENT_OK



"""Class for stream 06 function 08."""

from secsgem.secs.data_items import CEID, DATAID, DSID, DVNAME, DVVAL
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F08(SecsStreamFunction):
    """data transfer data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F08
        {
            DATAID: U1/U2/U4/U8/I1/I2/I4/I8/A
            CEID: U1/U2/U4/U8/I1/I2/I4/I8/A
            DS: [
                {
                    DSID: U1/U2/U4/U8/I1/I2/I4/I8/A
                    DV: [
                        {
                            DVNAME: U1/U2/U4/U8/I1/I2/I4/I8/A
                            DVVAL: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
                        }
                        ...
                    ]
                }
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F08({
        ...     "DATAID": 1,
        ...     "CEID": 1337,
        ...     "DS": [{
        ...         "DSID": 1000,
        ...         "DV": [
        ...             {"DVNAME": "VAR1", "DVVAL": "VAR"},
        ...             {"DVNAME": "VAR2", "DVVAL": secsgem.secs.variables.U4(100)}]}]})
        S6F8
          <L [3]
            <U1 1 >
            <U2 1337 >
            <L [1]
              <L [2]
                <U2 1000 >
                <L [2]
                  <L [2]
                    <A "VAR1">
                    <A "VAR">
                  >
                  <L [2]
                    <A "VAR2">
                    <U4 100 >
                  >
                >
              >
            >
          > .

    Data Items:
        - :class:`DATAID <secsgem.secs.data_items.DATAID>`
        - :class:`CEID <secsgem.secs.data_items.CEID>`
        - :class:`DSID <secsgem.secs.data_items.DSID>`
        - :class:`DVNAME <secsgem.secs.data_items.DVNAME>`
        - :class:`DVVAL <secsgem.secs.data_items.DVVAL>`

    """

    _stream = 6
    _function = 8

    _data_format = [
        DATAID,
        CEID,
        [
            [
                "DS",
                DSID,
                [
                    [
                        "DV",
                        DVNAME,
                        DVVAL
                    ]
                ]
            ]
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Class for stream 09 function 05."""

from secsgem.secs.data_items import MHEAD
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS09F05(SecsStreamFunction):
    """unrecognized function type.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS09F05
        MHEAD: B[10]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS09F05("HEADERDATA")
        S9F5
          <B 0x48 0x45 0x41 0x44 0x45 0x52 0x44 0x41 0x54 0x41> .

    Data Items:
        - :class:`MHEAD <secsgem.secs.data_items.MHEAD>`

    """

    _stream = 9
    _function = 5

    _data_format = MHEAD

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = False



"""Wrapper for GEM alarm."""
from __future__ import annotations

import secsgem.secs


class Alarm:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """Alarm definition."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        alid: str | int,
        name: str,
        text: str,
        code: int,
        ce_on: str | int,
        ce_off: str | int,
        **kwargs
    ):
        """Initialize an alarm.

        You can manually set the secs-type of the id with the 'id_type' keyword argument.

        Args:
            alid: ID of the alarm
            name: long name of the alarm
            text: alarm text
            code: alarm code
            ce_on: collection event for alarm set
            ce_off: collection event for alarm cleared
            **kwargs: additional attributes for object

        """
        self.alid = alid
        self.name = name
        self.text = text
        self.code = code
        self.ce_on = ce_on
        self.ce_off = ce_off
        self.enabled = False
        self.set = False

        self.id_type: type[secsgem.secs.variables.Base]

        if isinstance(self.alid, int):
            self.id_type = secsgem.secs.variables.U4
        else:
            self.id_type = secsgem.secs.variables.String

        for key, value in kwargs.items():
            setattr(self, key, value)



"""SOFTREV data item."""
from .. import variables
from .base import DataItemBase


class SOFTREV(DataItemBase):
    """Software revision.

    :Type: :class:`String <secsgem.secs.variables.String>`
    :Length: 20

    """

    __type__ = variables.String
    __count__ = 20



"""Class for stream 05 function 11."""

from secsgem.secs.data_items import EXID, EXMESSAGE, EXTYPE, TIMESTAMP
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS05F11(SecsStreamFunction):
    """exception clear - notify.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F11
        {
            TIMESTAMP: A[32]
            EXID: A[20]
            EXTYPE: A
            EXMESSAGE: A
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS05F11({
        ...     "TIMESTAMP": "161006221500",
        ...     "EXID": "EX123",
        ...     "EXTYPE": "ALARM",
        ...     "EXMESSAGE": "Exception"})
        S5F11
          <L [4]
            <A "161006221500">
            <A "EX123">
            <A "ALARM">
            <A "Exception">
          > .

    Data Items:
        - :class:`TIMESTAMP <secsgem.secs.data_items.TIMESTAMP>`
        - :class:`EXID <secsgem.secs.data_items.EXID>`
        - :class:`EXTYPE <secsgem.secs.data_items.EXTYPE>`
        - :class:`EXMESSAGE <secsgem.secs.data_items.EXMESSAGE>`

    """

    _stream = 5
    _function = 11

    _data_format = [
        TIMESTAMP,
        EXID,
        EXTYPE,
        EXMESSAGE
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = True
    _is_reply_required = False

    _is_multi_block = False



"""Class for stream 06 function 22."""

from secsgem.secs.data_items import VID, V
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS06F22(SecsStreamFunction):
    """annotated individual report data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F22
        [
            {
                VID: U1/U2/U4/U8/I1/I2/I4/I8/A
                V: L/BOOLEAN/U1/U2/U4/U8/I1/I2/I4/I8/F4/F8/A/B
            }
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS06F22([{"VID": "VID1", "V": "ASD"}, {"VID": 2, "V": 1337}])
        S6F22
          <L [2]
            <L [2]
              <A "VID1">
              <A "ASD">
            >
            <L [2]
              <U1 2 >
              <U2 1337 >
            >
          > .

    Data Items:
        - :class:`VID <secsgem.secs.data_items.VID>`
        - :class:`V <secsgem.secs.data_items.V>`

    """

    _stream = 6
    _function = 22

    _data_format = [
        [
            VID,
            V
        ]
    ]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""Header for the hsms messages."""
from __future__ import annotations

import enum
import struct
import typing

import secsgem.common


class HsmsSType(enum.Enum):
    """HSMS SType enum."""

    DATA_MESSAGE = 0
    SELECT_REQ = 1
    SELECT_RSP = 2
    DESELECT_REQ = 3
    DESELECT_RSP = 4
    LINKTEST_REQ = 5
    LINKTEST_RSP = 6
    REJECT_REQ = 7
    SEPARATE_REQ = 9

    @classmethod
    def names(cls) -> dict[HsmsSType, str]:
        """Get the names associated with the scode.

        Returns:
            dictionary of names associated with enum values

        """
        return {
            cls.DATA_MESSAGE: "Data.msg",
            cls.SELECT_REQ: "Select.req",
            cls.SELECT_RSP: "Select.rsp",
            cls.DESELECT_REQ: "Deselect.req",
            cls.DESELECT_RSP: "Deselect.rsp",
            cls.LINKTEST_REQ: "Linktest.req",
            cls.LINKTEST_RSP: "Linktest.rsp",
            cls.REJECT_REQ: "Reject.req",
            cls.SEPARATE_REQ: "Separate.req",
        }

    @property
    def text(self) -> str:
        """Get the text for the item."""
        return self.names()[self]


class HsmsHeader(secsgem.common.Header):
    """Generic HSMS header.

    Base for different specific headers
    """

    length = 10

    def __init__(  # pylint: disable=too-many-arguments
            self,
            system: int,
            session_id: int,
            stream: int = 0,
            function: int = 0,
            requires_response: bool = False,
            p_type: int = 0x00,
            s_type: HsmsSType = HsmsSType.SELECT_REQ):
        """Initialize a hsms header.

        Args:
            system: message ID
            session_id: device / session ID
            stream: stream
            function: function
            requires_response: is response required
            p_type: P-Type
            s_type: S-Type

        Example:
            >>> import secsgem.hsms
            >>>
            >>> secsgem.hsms.HsmsHeader(3, 100)
            HsmsHeader({session_id:0x0064, stream:00, function:00, p_type:0x00, s_type:0x01, system:0x00000003, \
require_response:False})

        """
        super().__init__(system, session_id, stream, function, requires_response)
        self._p_type = p_type
        self._s_type = s_type

    def __str__(self) -> str:
        """Generate string representation for an object of this class."""
        return (
            f"{{session_id:0x{self.session_id:04x}, "
            f"stream:{self.stream:02d}, "
            f"function:{self.function:02d}, "
            f"p_type:0x{self.p_type:02x}, "
            f"s_type:0x{self.s_type.value:02x}, "
            f"system:0x{self.system:08x}, "
            f"require_response:{self.require_response!r}}}"
        )

    def __repr__(self) -> str:
        """Generate textual representation for an object of this class."""
        return f"{self.__class__.__name__}({self.__str__()})"

    @property
    def p_type(self) -> int:
        """Get P-type."""
        return self._p_type

    @property
    def s_type(self) -> HsmsSType:
        """Get S-type."""
        return self._s_type

    @property
    def _as_dictionary(self) -> dict[str, typing.Any]:
        """Get the data as dictionary.

        Returns:
            Header data as dictionary.

        """
        return {
            "system": self._system,
            "session_id": self._session_id,
            "stream": self._stream,
            "function": self._function,
            "requires_response": self._require_response,
            "p_type": self._p_type,
            "s_type": self._s_type,
        }

    def encode(self) -> bytes:
        """Encode header to hsms message.

        Returns:
            encoded header

        Example:
            >>> import secsgem.hsms
            >>> import secsgem.common
            >>>
            >>> header = secsgem.hsms.HsmsLinktestReqHeader(2)
            >>> secsgem.common.format_hex(header.encode())
            'ff:ff:00:00:00:05:00:00:00:02'

        """
        header_stream = self.stream
        if self.require_response:
            header_stream |= 0b10000000

        return struct.pack(
            ">HBBBBL",
            self.session_id,
            header_stream,
            self.function,
            self.p_type,
            self.s_type.value,
            self.system,
        )

    @classmethod
    def decode(cls, data: bytes) -> HsmsHeader:
        """Decode data to HsmsHeader object.

        Args:
            data: data to decode

        Returns:
            new header object

        """
        res = struct.unpack(">HBBBBL", data)

        return HsmsHeader(
            res[5],
            res[0],
            res[1] & 0b01111111,
            res[2],
            (((res[1] & 0b10000000) >> 7) == 1),
            res[3],
            HsmsSType(res[4]),
        )



"""Class for stream 07 function 01."""

from secsgem.secs.data_items import LENGTH, PPID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F01(SecsStreamFunction):
    """process program load - inquire.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F01
        {
            PPID: A/B[120]
            LENGTH: U1/U2/U4/U8/I1/I2/I4/I8
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F01({"PPID": "program", "LENGTH": 4})
        S7F1 W
          <L [2]
            <A "program">
            <U1 4 >
          > .

    Data Items:
        - :class:`PPID <secsgem.secs.data_items.PPID>`
        - :class:`LENGTH <secsgem.secs.data_items.LENGTH>`

    """

    _stream = 7
    _function = 1

    _data_format = [
        PPID,
        LENGTH
    ]

    _to_host = True
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 02 function 37."""

from secsgem.secs.data_items import CEED, CEID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS02F37(SecsStreamFunction):
    """en-/disable event report.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F37
        {
            CEED: BOOLEAN[1]
            CEID: [
                DATA: U1/U2/U4/U8/I1/I2/I4/I8/A
                ...
            ]
        }

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS02F37({"CEED": True, "CEID": [1337]})
        S2F37 W
          <L [2]
            <BOOLEAN True >
            <L [1]
              <U2 1337 >
            >
          > .

    Data Items:
        - :class:`CEED <secsgem.secs.data_items.CEED>`
        - :class:`CEID <secsgem.secs.data_items.CEID>`

    """

    _stream = 2
    _function = 37

    _data_format = [
        CEED,
        [CEID]
    ]

    _to_host = False
    _to_equipment = True

    _has_reply = True
    _is_reply_required = True

    _is_multi_block = False



"""Class for stream 07 function 20."""

from secsgem.secs.data_items import PPID
from secsgem.secs.functions.base import SecsStreamFunction


class SecsS07F20(SecsStreamFunction):
    """current equipment process program - data.

    Args:
        value: parameters for this function (see example)

    Examples:
        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F20
        [
            PPID: A/B[120]
            ...
        ]

        >>> import secsgem.secs
        >>> secsgem.secs.functions.SecsS07F20(["program1", "program2"])
        S7F20
          <L [2]
            <A "program1">
            <A "program2">
          > .

    Data Items:
        - :class:`PPID <secsgem.secs.data_items.PPID>`

    """

    _stream = 7
    _function = 20

    _data_format = [PPID]

    _to_host = True
    _to_equipment = False

    _has_reply = False
    _is_reply_required = False

    _is_multi_block = True



"""ABS data item."""
from .. import variables
from .base import DataItemBase


class ABS(DataItemBase):
    """Any binary string.

    :Type: :class:`Binary <secsgem.secs.variables.Binary>`

    **Used In Function**
        - :class:`SecsS02F25 <secsgem.secs.functions.SecsS02F25>`
        - :class:`SecsS02F26 <secsgem.secs.functions.SecsS02F26>`

    """

    __type__ = variables.Binary



"""DATLC data item."""
from .. import variables
from .base import DataItemBase


class DATLC(DataItemBase):
    """Data location.

    :Type: :class:`U1 <secsgem.secs.variables.U1>`

    **Used In Function**
        - :class:`SecsS12F19 <secsgem.secs.functions.SecsS12F19>`

    """

    __type__ = variables.U1



"""Wrapper for GEM data value."""
from __future__ import annotations

import secsgem.secs


class DataValue:
    """Data value definition."""

    def __init__(self,
                 dvid: int | str,
                 name: str,
                 value_type: type[secsgem.secs.variables.Base],
                 use_callback: bool = True,
                 **kwargs):
        """Initialize a data value.

        You can manually set the secs-type of the id with the 'id_type' keyword argument.

        Custom parameters can be set with the keyword arguments,
        they will be passed to the GemEquipmentHandlers callback
        :func:`secsgem.gem.equipmenthandler.GemEquipmentHandler.on_dv_value_request`.

        If use_callbacks is disabled, you can set the value with the value property.

        Args:
            dvid: ID of the data value
            name: long name of the data value
            value_type: type of the data value
            use_callback: use the GemEquipmentHandler callbacks to get variable (True) or use internal value
            **kwargs: additional attributes for object

        """
        self._dvid = dvid
        self._name = name
        self._value_type = value_type
        self._use_callback = use_callback
        self.value = 0

        self._id_type: type[secsgem.secs.variables.Base]

        if isinstance(self._dvid, int):
            self._id_type = secsgem.secs.variables.U4
        else:
            self._id_type = secsgem.secs.variables.String

        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    def dvid(self) -> int | str:
        """Get the data value id."""
        return self._dvid

    @property
    def name(self) -> str:
        """Get the data value name."""
        return self._name

    @property
    def value_type(self) -> type[secsgem.secs.variables.Base]:
        """Get the data value type."""
        return self._value_type

    @property
    def use_callback(self) -> bool:
        """Get if data value uses callback."""
        return self._use_callback

    @property
    def id_type(self) -> type[secsgem.secs.variables.Base]:
        """Get the data value id type."""
        return self._id_type



"""PPID data item."""
from .. import variables
from .base import DataItemBase


class PPID(DataItemBase):
    """Process program ID.

    :Types:
       - :class:`String <secsgem.secs.variables.String>`
       - :class:`Binary <secsgem.secs.variables.Binary>`
    :Length: 120

    **Used In Function**
        - :class:`SecsS07F01 <secsgem.secs.functions.SecsS07F01>`
        - :class:`SecsS07F03 <secsgem.secs.functions.SecsS07F03>`
        - :class:`SecsS07F05 <secsgem.secs.functions.SecsS07F05>`
        - :class:`SecsS07F06 <secsgem.secs.functions.SecsS07F06>`
        - :class:`SecsS07F17 <secsgem.secs.functions.SecsS07F17>`
        - :class:`SecsS07F20 <secsgem.secs.functions.SecsS07F20>`
        - :class:`SecsS07F25 <secsgem.secs.functions.SecsS07F25>`

    """

    __type__ = variables.Dynamic
    __allowedtypes__ = [
        variables.String,
        variables.Binary
    ]
    __count__ = 120
